import numpy as np
import subprocess
import math
import gc
import os
import sys
from emmer.include.confidenceMapUtil import mapUtil, locscaleUtil, FDRutil

#--------------------------------------------------------------------------
def calculateConfidenceMap(em_map, apix, noiseBox, testProc, ecdf, lowPassFilter_resolution, method, window_size, locResMap,
						   meanMap, varMap, fdr, modelMap, stepSize, windowSizeLocScale, mpi):

	#*********************************************
	#******* this function calc. confMaps ********
	#*********************************************

	# get boxCoordinates
	if noiseBox is None:
		boxCoord = 0;
	else:
		boxCoord = noiseBox;

	# set test procdure
	if testProc is not None:
		testProc = testProc;
	else:
		testProc = 'rightSided';

	# set ECDF
	if ecdf:
		ECDF = 1;
	else:
		ECDF = 0;

	sizeMap = em_map.shape;

	if lowPassFilter_resolution is not None:
		frequencyMap = FDRutil.calculate_frequency_map(em_map);
		providedRes = apix/float(lowPassFilter_resolution);
		em_map = FDRutil.lowPassFilter(np.fft.rfftn(em_map), frequencyMap, providedRes, em_map.shape);

	# handle FDR correction procedure
	if method is not None:
		method = method;
	else:
		# default is Benjamini-Yekutieli
		method = 'BY';

	if window_size is not None:
		wn = window_size;
		wn = int(wn);
		if wn < 20:
			print("Provided window size is quite small. Please think about potential inaccuracies of your noise estimates!");
	else:
		wn = max(int(0.05 * sizeMap[0]), 10);

	if windowSizeLocScale is not None:
		wn_locscale = windowSizeLocScale;
		if window_size is None:
			wn = int(wn_locscale);
	else:
		wn_locscale = None;

	if stepSize is None:
		stepSize = 5;

	# generate a circular Mask
	sphere_radius = (np.max(sizeMap) // 2);
	circularMaskData = mapUtil.makeCircularMask(np.copy(em_map), sphere_radius);

	# plot locations of noise estimation
	if modelMap is None:
		pp = mapUtil.makeDiagnosticPlot(em_map, wn, False, boxCoord);
		print(
			"Generating diagnostic plot of noise estimation. Please have a look in 'diag_image.pdf' that the molecule does not fall into the region used for background noise estimation.");
		pp.savefig("diag_image.pdf");
		pp.close();
	else:
		pp = mapUtil.makeDiagnosticPlot(em_map, wn, True, boxCoord);
		print(
			"Generating diagnostic plot of noise estimation. Please have a look in 'diag_image.pdf' that the molecule does not fall into the region used for background noise estimation.");
		pp.savefig("diag_image.pdf");
		pp.close();


	# estimate noise statistics
	if ((locResMap is None) & (modelMap is None)):  # if no local Resolution map is given,don't do any filtration

		FDRutil.checkNormality(em_map, wn, boxCoord);
		mean, var, _ = FDRutil.estimateNoiseFromMap(em_map, wn, boxCoord);

		if varMap is not None:
			var = varMap;
		if meanMap is not None:
			mean = meanMap;

		if np.isscalar(mean) and np.isscalar(var):
			output = "Estimated noise statistics: mean: " + repr(mean) + " and variance: " + repr(var);
		else:
			output = "Using user provided noise statistics";

		print(output);

		locFiltMap = None;
		locScaleMap = None;

	elif (locResMap is not None) & (modelMap is None):  # do localFiltration and estimate statistics from this map

		FDRutil.checkNormality(em_map, wn, boxCoord);
		em_map, mean, var, ECDF = mapUtil.localFiltration(em_map, locResMap, apix, True, wn, boxCoord, ECDF);
		#locFiltMap = FDRutil.studentizeMap(em_map, mean, var);
		locFiltMap = em_map;
		locScaleMap = None;
	else:
		em_map, mean, var, ECDF = locscaleUtil.launch_amplitude_scaling(em_map, modelMap, apix, stepSize, wn_locscale, wn, method, locResMap, boxCoord, mpi, ECDF );
		#locScaleMap = FDRutil.studentizeMap(em_map, mean, var);
		locScaleMap = em_map;
		locFiltMap = None;

	# calculate the qMap
	qMap = FDRutil.calcQMap(em_map, mean, var, ECDF, wn, boxCoord, circularMaskData, method, testProc);

	# if a explicit thresholding is wished, do so
	if (method == 'BY') | (method == 'BH'):
		error = "FDR";	
	else:
		error = "FWER";
		
	if fdr is not None:

		fdr = fdr;

		# threshold the qMap
		binMap = FDRutil.binarizeMap(qMap, fdr);

		# apply the thresholded qMap to data
		maskedMap = np.multiply(binMap, em_map);
		minMapValue = np.min(maskedMap[np.nonzero(maskedMap)]);

		maskedMap = np.multiply(maskedMap, circularMaskData);

		if (locResMap is None) & (modelMap is None):  # if no local Resolution map is give, then give the correspoding threshold, not usefule with local filtration
			output = "Calculated map threshold: %.3f" %minMapValue + " at a " + error +  " of " + repr(fdr*100) + " %.";
			print(output);
	else:
		# threshold the qMap
		binMap1 = FDRutil.binarizeMap(qMap, 0.01);
		binMap001 = FDRutil.binarizeMap(qMap, 0.0001);
		
		if (locResMap is None) & (modelMap is None):  # if no local Resolution map is give, then give the correspoding threshold, not usefule with local filtration
			# apply the thresholded qMap to data
			maskedMap1 = np.multiply(binMap1, np.copy(em_map));
			maskedMap001 = np.multiply(binMap001, np.copy(em_map));
			minMapValue1 = np.min(maskedMap1[np.nonzero(maskedMap1)]);
			minMapValue001 = np.min(maskedMap001[np.nonzero(maskedMap001)]);	
			output = "Calculated map threshold: %.3f" %minMapValue1 + " at a " + error + " of " + repr(1) +  " %.";
			print(output);
			output = "Calculated map threshold: %.3f" %minMapValue001 + " at a " + error + " of " + repr(0.01) + " %.";
			print(output);
		"""elif (locResMap is not None) & (modelMap is None):
			# apply the thresholded qMap to data
			maskedMap = np.multiply(binMap, np.copy(locFiltMap));
			minMapValue = np.min(maskedMap[np.nonzero(maskedMap)]);
			output = "Calculated map threshold: %.3f"  %minMapValue + " at a " + error + " of " + repr(fdr*100) + "%.";
			print(output);
		elif (locResMap is None) & (modelMap is not None):
			# apply the thresholded qMap to data
			maskedMap = np.multiply(binMap, np.copy(locScaleMap));
			minMapValue = np.min(maskedMap[np.nonzero(maskedMap)]);
			output = "Calculated map threshold: %.3f" %minMapValue + " at a " + error + " of " + repr(fdr*100) + "%.";
			print(output);
		"""	
		binMap = None;
		maskedMap = None;

	# invert qMap for visualization tools
	confidenceMap = np.subtract(1.0, qMap);

	# apply lowpass-filtered mask to maps
	confidenceMap = np.multiply(confidenceMap, circularMaskData);

	return confidenceMap, locFiltMap, locScaleMap, binMap, maskedMap;



# -*- coding: utf-8 -*-
"""

"""
import numpy as np
import gemmi
import mrcfile
import matplotlib.pyplot as plt

#%%


def plot_mapContour_model(threshold=0, modelfile="", gemmi_strct="", mapfile="", mrcfile_map="", projection_orientation='z'):
    """
    Function for plotting density map as contour plot with model overlay.
    

    Parameters
    ----------
    threshold : float, optional
        threshold value for plotting. The default is 0.
    modelfile : str, optional
        pdb file to use as model structure. Either modelfile or gemmi_strct must be specified.
    gemmi_strct : gemmi.Structure, optional
        Gemmi structure object containing at least 1 model. Either modelfile or gemmi_strct must be specified.
    mapfile : str, optional
        filename of the target map to be plotted. Either mapfile or mrcfile_map must be specified.
    mrcfile_map : MrcFile, optional
        mrcFile object of the target map.  Either mapfile or mrcfile_map must be specified.
    projection_orientation : str, optional
        axis along which to project the target map, can be 'x', 'y' or 'z'. The default is 'z'.


    Returns
    -------
    None.

    """
    
    if mapfile!="":
        target_map = mrcfile.open(mapfile)
    elif mrcfile_map!="":
        target_map = mrcfile_map
    else:
        print("must specify target map as either path to .mrc file or as mrcfile object")
        return 0
    voxelsize = target_map.header.cella.x / target_map.header.nx
    
    if projection_orientation=='x':
        projection_axis = 0
        plot_axes = [1,2]
    elif projection_orientation=='y':
        projection_axis = 1
        plot_axes = [0,2]
    elif projection_orientation=='z':
        projection_axis = 2
        plot_axes = [0,1]
    else:
        print("please specify valid projection orientation ('x','y','z')")
        return 0
    
    if modelfile!="":
        gemmi_strct = gemmi.read_structure(modelfile)
    elif gemmi_strct=="":
        print("must specify either path to .pdb file or a gemmi.Structure object")
        return 0
        
    T = np.copy(target_map.data) # mrcfile.data object is read only
    T[T<threshold] = np.nan
    
    x = np.linspace(0, target_map.header.cella.x, target_map.header.nx)
    XX,YY = np.meshgrid(x,x)
    
    # prepare plot figure
    projection = np.nanmean(T, axis=projection_axis)
    # projection = np.mean(target_map.data,axis=2)
    fig, ax = plt.subplots()
    # im = ax.imshow(projection, interpolation='spline16')
    cntr = ax.contourf(XX,YY,projection, levels = 32, alpha=0.8)
    cbar = fig.colorbar(im)
    ax.axis('off')
    ax.set_aspect('equal')
    
    # get number of backbone atoms
    numAtoms = 0
    for chn in gemmi_strct[0]:
        for res in chn:
            for atm in res:
                if atm.name == 'C' or atm.name=='N' or atm.name=='CA':
                    numAtoms += 1
    
    # find positions of all backbone atoms
    plot_pos = np.zeros((numAtoms, 3))
    atmidx = 0
    for chn in gemmi_strct[0]:
        for res in chn:
            for atm in res:
                if atm.name == 'C' or atm.name=='N' or atm.name=='CA':
                    if projection_orientation=='x':
                        plot_pos[atmidx,:] = np.array([atm.pos.z, atm.pos.x, atm.pos.y])
                    if projection_orientation=='z':
                        plot_pos[atmidx,:] = np.array([atm.pos.y, atm.pos.z, atm.pos.x])
                    if projection_orientation=='y':
                        plot_pos[atmidx,:] = np.array([atm.pos.x, atm.pos.y, atm.pos.z])      
                    atmidx+=1

    ax.plot(plot_pos[:,plot_axes[0]], plot_pos[:,plot_axes[1]],
            marker='o',
            markerfacecolor='red',
            markeredgecolor='black',
            markersize=3,
            linewidth=1.5,
            color='red')    
    
        


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 14 10:06:12 2021
"""

import numpy as np
import pypdb
import gemmi
from emmer.ndimage.sharpen.sharpen_maps import sharpen_maps
from emmer.ndimage.map_utils import convert_to_tuple
from emmer.pdb.pdb_utils import detect_pdb_input
from scipy.ndimage import center_of_mass
import warnings

def eval_size_apix_unitcell(pdb_structure, size, apix, unitcell, verbose):
    # out of (size, apix, unitcell) at least 2 parameters need to be given. If all three are given, the parameters must be in agreement. If only one of (size, apix) is given, the function will use whatever unitcell is in the gemmi structure
    
    # case 1: the user specifies (size, apix)
    if size is not None and apix is not None and unitcell is None:
        if verbose:
            print("Map size and pixelsize are given as inputs. Unitcell present in input structure will be ignored.")
        size_gemmi = size
        apix_gemmi = convert_to_tuple(apix, num_dims=3) # If apix is scalar, then the function returns a tuple of len 3
        # determine unknown
        unitcell_gemmi = gemmi.UnitCell(size[0]*apix_gemmi[0], size[1]*apix_gemmi[1], size[2]*apix_gemmi[2], 90, 90, 90)
        
    # case 2: the user specifies (size, unitcell)
    elif size is not None and apix is None and unitcell is not None:
        if verbose:
            print("Map size and unitcell are given as inputs. Unitcell present in input structure will be ignored.")
        unitcell_gemmi = unitcell
        size_gemmi = size
        # determine unknown
        apix_gemmi = (unitcell_gemmi.a/size_gemmi[0],unitcell_gemmi.b/size_gemmi[1],unitcell_gemmi.c/size_gemmi[2])
    
    # case 3: the user specifies (apix, unitcell)
    elif size is None and apix is not None and unitcell is not None:
        if verbose:
            print("Pixel size and unitcell are given as inputs. Unitcell present in input structure will be ignored.")
        unitcell_gemmi = unitcell
        apix_gemmi = convert_to_tuple(apix,num_dims=3)
        # determine unknown
        size_gemmi = [int(round(unitcell.a/apix_gemmi[0])), int(round(unitcell.b/apix_gemmi[1])), int(round(unitcell.c/apix_gemmi[2]))]
        
    # case 4: the user specifies only size
    elif size is not None and unitcell is None:  # in this case, program will use whatever unitcell is present in the gemmi structure
        if verbose:
            print("Map size is given as input. Unitcell present in the structure will be used")
        size_gemmi = size
        unitcell_gemmi = pdb_structure.cell
        # determine unknown
        apix_gemmi = (pdb_structure.cell.a/size_gemmi[0],pdb_structure.cell.b/size_gemmi[1],pdb_structure.cell.c/size_gemmi[2])
        
    # case 5: the user specifies only apix
    elif apix is not None and unitcell is None:
        if verbose:
            print("Pixel size is given as input. Unitcell present in the structure will be used")
        apix_gemmi = apix
        unitcell_gemmi = pdb_structure.cell
        # determine unknown
        size_gemmi = [int(round(unitcell_gemmi.a/apix_gemmi[0])), int(round(unitcell_gemmi.b/apix_gemmi[1])), int(round(unitcell_gemmi.c/apix_gemmi[2]))]

    # case 6: the used specifies (size, apix, unitcell). Raises warning if not in agreement and then proceeds to use the size and the apix
    elif size is not None and apix is not None and unitcell is not None:
        if verbose:
            print("Map size, pixel size and unitcell are given as inputs. Unitcell present in input structure will be ignored. Checking if parameters are self-consistent")
        size_gemmi = size
        apix_gemmi = convert_to_tuple(apix,num_dims=3)
        unitcell_gemmi = unitcell
        if int(round(unitcell_gemmi.a / apix_gemmi[0])) != size_gemmi[0] or int(round(unitcell_gemmi.b / apix_gemmi[1])) or int(round(unitcell_gemmi.c / apix_gemmi[2])):
            warnings.warn("size, pixel size and unitcell parameters are inconsistent. Proceeding with only size and pixel size")
            unitcell_gemmi = gemmi.UnitCell(size[0]*apix_gemmi[0], size[1]*apix_gemmi[1], size[2]*apix_gemmi[2], 90, 90, 90)
        
    return size_gemmi, apix_gemmi, unitcell_gemmi
        

def convert_pdb_to_map(input_pdb=None, unitcell=None, size=None, apix=None, return_grid=False, verbose=False, 
            mdlidx=0,align_output=True,set_refmac_blur=True,set_unblur=True, blur=0):
    '''
    Cleaner function to convert a gemmi_structure to EM map. Make sure the input structure, or the pdb in the 
    path you input are correct. Common check include: 
        a) center of mass of the gemmi structure should be roughly in the center of unitcell (important)
        b) Remove waters 
        c) Check if atomic bfactors make sense
    
    Note: if you use a single value for apix then the program will assume same voxelsize in all dimensions

    Parameters
    ----------
    input_pdb : str or gemmi.Structure, required
        Input can be either 
        a) string: either (i) pdb path, for ex: "path/to/pdb.pdb" or 
                          (ii)pdb_id "3j5p"
        b) gemmi.Structure()

    unitcell : gemmi.UnitCell, 
        
    size : tuple, 
        Expected map shape
    apix : float, optional
        Expected voxelsize
    Either two of the three (unitcell/size/apix) allowed. In case only size is given as input, unitcell is taken 
    from gemmi structure
    return_grid : bool, optional
        
    verbose : TYPE, optional
        The default is False.
    mdlidx : int
        If gemmi.Structure has multiple models this index tells which model to select. Default is zero
    align_output : bool
        If selected, this transforms to output to align according to mrcfile convenctions 
        Transformation: flip axis: 2 then rotate in plane (2,0) by angle 90
        

    Returns
    -------
    if return_grid is set True: emmap (numpy.ndarray), grid (gemmi.FloatGrid)
    
    Else, only emmap is returned

    '''
    pdb_structure = detect_pdb_input(input_pdb)
    size_gemmi, apix_gemmi, unitcell_gemmi = eval_size_apix_unitcell(pdb_structure, size, apix, unitcell, verbose)
    pdb_structure.cell = unitcell_gemmi
    
    reporter = {}
    reporter['pdb_struct_name'] =    pdb_structure.name
    reporter['unit_cell_exp'] =      pdb_structure.cell.parameters
    reporter['shape_exp'] = size_gemmi
    reporter['apix_exp'] = apix_gemmi
    reporter['com_pdb'] = np.array(pdb_structure[mdlidx].calculate_center_of_mass().tolist())
    if verbose:   # Make a check: 
        float_formatter = "{:.4f}".format
        np.set_printoptions(formatter={'float_kind':float_formatter})
        print("Simulating EM-map using: \t ", reporter['pdb_struct_name'], " \nwith the following properties: \n")
        print("Unit-cell (A*A*A): \t\t", reporter['unit_cell_exp'])
        print("Expected shape: \t\t", reporter['shape_exp'])
        print("Expected voxelsize (A*A*A): \t\t", reporter['apix_exp'])
        print("Center of mass: (A*A*A): \t\t",reporter['com_pdb'] )
    
    ## At this point, the values for size, pixelsize the unit cell are set 
    ## we should use the variable size_gemmi in density calculator

    if mdlidx > len(pdb_structure):  # Select model index from the pdb_structure
        print("selected model number not in pdb")
        return 0
    model = pdb_structure[mdlidx]
    dencalc = gemmi.DensityCalculatorE()
    dencalc.d_min = apix_gemmi[0] / 2
    dencalc.rate = 1
    if blur > 0:
        dencalc.blur=blur
    if set_refmac_blur:
        dencalc.set_refmac_compatible_blur(model)
        inv_d2 = dencalc.blur
        if verbose:
            print("Setting a blur of {:.2f}".format(dencalc.blur))
            
    dencalc.set_grid_cell_and_spacegroup(pdb_structure)
    try:
        dencalc.grid.set_size(*size_gemmi)
        dencalc.add_model_density_to_grid(model)
    except:
        dencalc.put_model_density_on_grid(model)
        print("WARNING: You are using an older version of gemmi. " +
              "The map size of the pdb generated map might not be as expected")
        print("--> Please update your gemmi package to the newest version")

    emmap = np.array(dencalc.grid,copy=False)
    
    if set_refmac_blur:
        if verbose: 
            print("Applying a unblur for the sampled density equal to: {:.2f}".format(-inv_d2))
        emmap = sharpen_maps(emmap, apix=apix_gemmi[0], sharpening_bfactor=-inv_d2)
        
    
    if align_output:
        from scipy.ndimage import rotate
        emmap_flipped = np.flip(emmap,axis=2)
        emmap_rotated = rotate(emmap_flipped, angle=90, axes=(2,0))
        emmap = emmap_rotated
    
    reporter['shape_final'] = emmap.shape
    reporter['apix_final'] = np.array(dencalc.grid.spacing)
    reporter['com_map'] = np.array(center_of_mass(abs(emmap)))*reporter['apix_final']
    if verbose: ## Check output if it matches expectation
        print("\nMap simulated! Final parameters:")
        print("Emmap shape \t\t",reporter['shape_final'])
        print("Grid spacing (A*A*A): \t\t", reporter['apix_final'])
        print("Center of mass of Emmap: (A*A*A)\t", reporter['com_map'])
    
    if return_grid:
        return emmap, dencalc.grid
    
    else:
        return emmap
    


## Script to perform low pass filtering of cryo EM maps
import numpy as np
def low_pass_filter(im, cutoff, apix):

    """
    Returns a low-pass filter image from a tanh filter.
    """
    import numpy as np

    from emmer.ndimage.filter.filter_utils import calculate_fourier_frequencies, tanh_filter
    
    im_freq     = calculate_fourier_frequencies(im, apix=apix)
    im_filter   = tanh_filter(im_freq, cutoff);
    im_fft      = np.fft.rfftn(im)
    im_fft_filtered = im_fft * im_filter
    im_filtered = np.fft.irfftn(im_fft_filtered)
    return im_filtered


def convert_pdb_to_mrc_position(pdb_position, apix):
    '''
    Convert the real units of positions into indices for the emmap. 
    Note: returns in (Z,Y,X) format
    

    Parameters
    ----------
    pdb_position : list
        list of xyz positions (Angstorm)
    apix : float
        Pixel size 

    Returns
    -------
    mrc_position : list
        List of ZYX positions (index positions)

    '''
    mrc_position = []
    
    for pos in pdb_position:
        [x,y,z] = pos
        int_x, int_y, int_z = int(round(x/apix)), int(round(y/apix)), int(round(z/apix))
        mrc_position.append([int_z, int_y, int_x])
        
    return mrc_position



# -*- coding: utf-8 -*-
"""
Created on 25 April 2022

@author: alok
"""

import unittest
import gemmi
import mrcfile
import numpy as np
import os

class test_pdb_shake(unittest.TestCase):
    def setUp(self):
        import pickle
        import emmer
        
        emmer_module_path = os.path.dirname(emmer.__file__)
        emmer_install_path = emmer_module_path
        self.emmer_install_path = emmer_install_path
        os.sys.path.insert(0, self.emmer_install_path)
        
        self.data_folder = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','data') 
        self.emmap_path = os.path.join(self.data_folder,'emd5778_map.mrc')
        self.mask_path = os.path.join(self.data_folder,'emd5778_mask.mrc')
        self.pdb_path = os.path.join(self.data_folder,'pdb3j5p.pdb')
        self.modmap_path = os.path.join(self.data_folder,'pdb3j5p_map.mrc')
        
        verify_output_file = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','verify','ndimage_outputs.pickle') 
        with open(verify_output_file, "rb") as outfile:
            self.validate = pickle.load(outfile)
        
    def copy_files(self, file_path, tempDir):
        from subprocess import run
        run(["cp",file_path,tempDir])
        if os.path.exists(os.path.join(tempDir, os.path.basename(file_path))):               
            return os.path.join(tempDir, os.path.basename(file_path))
            
        else:
            raise UserWarning("Could not copy {} to {}".format(path,tempDir))
        
    def test_shake_pdb(self):
        from emmer.pdb.pdb_tools.perturb_within_mask import perturb_within_mask
        from emmer.pdb.pdb_tools.compute_rmsd_two_pdb import compute_rmsd_two_pdb
        from emmer.ndimage.mask.get_atomic_model_mask import get_atomic_model_mask
        from emmer.ndimage.map_utils import save_as_mrc

        pdb_path = self.pdb_path
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory
        with TemporaryDirectory() as tmpdir:
            copied_pdb_path = self.copy_files(pdb_path, tmpdir)
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)

            ## get atomic model mask
            emmap = mrcfile.open(copied_sample_emmap_path)
            model_mask = get_atomic_model_mask(emmap=emmap.data, apix=emmap.voxel_size.x, pdb_path=copied_pdb_path, dilation_radius=3, softening_parameter=5)
            save_as_mrc(model_mask, os.path.join(tmpdir,'model_mask.mrc'),apix=1.2156)
            model_mask_path = os.path.join(tmpdir,'model_mask.mrc')
            ## shake pdb within mask
            for trial in range(5):
                rmsd_mag = np.random.uniform(low=5, high=25)
                shaken_pdb = perturb_within_mask(copied_pdb_path, model_mask_path, rmsd_magnitude=rmsd_mag)

                RMSD_shake = compute_rmsd_two_pdb(copied_pdb_path, shaken_pdb)

                percentage_rmsd_difference = abs(RMSD_shake-rmsd_mag)/rmsd_mag*100
                
                self.assertLess(percentage_rmsd_difference, 15)

        

        
        

       
        

      


      


## Script to return the threshold that gives the closest volume to a reference volume

def volume_matching_threshold(emmap, apix, reference_volume, num_bins=100, min_threshold=None, max_threshold=None):
    import numpy as np
    
    if min_threshold is None:
        min_threshold = 0
    
    if max_threshold is None:
        max_threshold = 0.97 * emmap.max()
        
    threshold_bins = np.linspace(min_threshold, max_threshold, num=num_bins)
    volumes = []
    for threshold in threshold_bins:
        binarised_map = (emmap>=threshold).astype(np.int_)
        sum_of_voxels = binarised_map.sum()
        volume_real_units = sum_of_voxels * apix**3
        volumes.append(volume_real_units)
    
    volumes = np.array(volumes)
    index_min_delta = np.argmin(abs(volumes-reference_volume))
    matching_threshold = threshold_bins[index_min_delta]
    
    return matching_threshold



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 25 April 2022

@author: alok
"""

import unittest
import gemmi
import mrcfile
import numpy as np
import os

class test_map_tools(unittest.TestCase):
    def setUp(self):
        import pickle
        import emmer
        emmer_module_path = os.path.dirname(emmer.__file__)
        emmer_install_path = emmer_module_path
        self.emmer_install_path = emmer_install_path
        os.sys.path.insert(0, self.emmer_install_path)
        
        self.data_folder = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','data') 
        self.emmap_path = os.path.join(self.data_folder,'emd5778_map.mrc')
        self.mask_path = os.path.join(self.data_folder,'emd5778_mask.mrc')
        self.pdb_path = os.path.join(self.data_folder,'pdb3j5p.pdb')
        self.modmap_path = os.path.join(self.data_folder,'pdb3j5p_map.mrc')
        
        verify_output_file = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','verify','ndimage_outputs.pickle') 
        with open(verify_output_file, "rb") as outfile:
            self.validate = pickle.load(outfile)
        
    def copy_files(self, file_path, tempDir):
        from subprocess import run
        run(["cp",file_path,tempDir])
        if os.path.exists(os.path.join(tempDir, os.path.basename(file_path))):               
            return os.path.join(tempDir, os.path.basename(file_path))
            
        else:
            raise UserWarning("Could not copy {} to {}".format(path,tempDir))
        


    def test_atomic_model_mask(self):
        ## shake pdb within mask at 2A resolution
        from emmer.pdb.pdb_tools.perturb_within_mask import perturb_within_mask
        from emmer.pdb.pdb_utils import get_all_atomic_positions
        from emmer.ndimage.mask.get_atomic_model_mask import get_atomic_model_mask
        from emmer.pdb.convert.convert_pdb_to_mrc_position import convert_pdb_to_mrc_position
        from emmer.ndimage.map_utils import load_map
        pdb_path = self.pdb_path
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_pdb_path = self.copy_files(pdb_path, tmpdir)
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap, apix = load_map(copied_sample_emmap_path)

            ## get atomic model mask
            model_mask = get_atomic_model_mask(emmap=emmap, apix=1.2156,pdb_path=copied_pdb_path, dilation_radius=3, softening_parameter=5)
            binarised = (model_mask > 0.5).astype(np.int_)
            pdb_positions = get_all_atomic_positions(copied_pdb_path)
            mrc_positions = convert_pdb_to_mrc_position(pdb_positions, apix=1.2156)

            count = 0
            for mrc in mrc_positions:
                count += binarised[mrc[0],mrc[1],mrc[2]]
            
            num_atoms = len(pdb_positions)
            capture = count/num_atoms * 100
     
            self.assertTrue(capture > 99)


    def test_trim_map_between_residues(self):
        ## shake pdb within mask at 2A resolution
        from emmer.ndimage.trim_map_between_residues import trim_map_between_residues
        from emmer.ndimage.map_utils import load_map, save_as_mrc
        
        pdb_path = self.pdb_path
        sample_emmap_path = self.emmap_path


        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_pdb_path = self.copy_files(pdb_path, tmpdir)
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap, _ = load_map(copied_sample_emmap_path)

            ## crop map between residues 
            chain_name = "B"
            residue_range = [378,383]

            cropped_map = trim_map_between_residues(copied_sample_emmap_path, copied_pdb_path, chain_name, residue_range)

            cropped_map_target = self.validate['map_tools_output']['cropped_map']

            self.assertTrue(np.allclose(cropped_map, cropped_map_target))

    def test_confidence_map(self):
        from emmer.ndimage.map_tools import measure_mask_parameters
        from emmer.ndimage.mask.compute_confidence_map import compute_confidence_map
        from emmer.ndimage.map_utils import load_map
        from emmer.ndimage.compute_real_space_correlation import compute_real_space_correlation
        sample_emmap_path = self.emmap_path
        sample_mask_path = self.mask_path
        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            copied_sample_mask_path = self.copy_files(sample_mask_path, tmpdir)
            emmap, _ = load_map(copied_sample_emmap_path)

            ## compute confidence map
            confidence_map = compute_confidence_map(emmap, apix=1.2156, window_size=25, fdr=0.01)
           
            confidence_map_target,_ = load_map(copied_sample_mask_path)
            rscc = compute_real_space_correlation(confidence_map, confidence_map_target)

            self.assertTrue(rscc > 0.8)            

            mask_parameters = measure_mask_parameters(mask=confidence_map, apix=1.2156, detailed_report=True)
            
            self.assertTrue(mask_parameters[2]==41926)

    def test_resample_map(self):
        from emmer.ndimage.map_tools import resample_map
        from emmer.ndimage.map_utils import load_map
        
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap, apix = load_map(copied_sample_emmap_path)

            resampled_map = resample_map(emmap, apix=1.2156, apix_new=1.0)



            self.assertTrue(resampled_map.shape == (311,311,311))
            self.assertAlmostEqual(resampled_map.mean(),emmap.mean(), delta=1e-5)
            self.assertAlmostEqual(resampled_map.std(),emmap.std(), delta=1e-4)
    
    def test_resample_image(self):
        from emmer.ndimage.map_tools import resample_image
        from emmer.ndimage.map_utils import load_map
        
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap, apix = load_map(copied_sample_emmap_path)

            resampled_map = resample_image(emmap, apix=1.2156, apix_new=1.0)



            self.assertTrue(resampled_map.shape == (311,311,311))
            self.assertAlmostEqual(resampled_map.mean(),emmap.mean(), delta=1e-4)
            self.assertAlmostEqual(resampled_map.std(),emmap.std(), delta=1e-3)
    
    def test_add_half_maps(self):
        from emmer.ndimage.average_half_maps import average_half_maps
        from emmer.ndimage.map_utils import load_map, save_as_mrc
        import os

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            
            random_map_1 = np.random.uniform(0,1,(256,256,256))
            random_map_2 = np.random.uniform(0,1,(256,256,256))

            save_as_mrc(random_map_1, os.path.join(tmpdir,'temp_1.mrc'), apix=1)
            save_as_mrc(random_map_2, os.path.join(tmpdir,'temp_2.mrc'), apix=1)

            full_map_path = average_half_maps(os.path.join(tmpdir,'temp_1.mrc'), os.path.join(tmpdir,'temp_2.mrc'), output_filename=os.path.join(tmpdir,'temp_full.mrc'))
            
            full_map, apix = load_map(full_map_path)
            self.assertTrue(os.path.exists(full_map_path))
            self.assertTrue(full_map.shape == random_map_1.shape)
            self.assertAlmostEqual(full_map.mean(),np.mean(np.array([random_map_1.mean(),random_map_2.mean()])), delta=1e-5)

            


            


        
        



import numpy as np
def sum_of_psd_2d(images):
    from emmer.ndimage.spectrum.spectrum_utils import psd2D
    psd2D_sum = None
    for image in images:
        if psd2D_sum is None:
            psd2D_sum = psd2D(image)
        else:
            psd2D_sum += psd2D(image)
    return np.array(psd2D_sum)





## Script to calculate phase correlation curve

def calculate_phase_correlation_maps(input_map_1, input_map_2):
    '''
    Wrapper to calculate FSC curve from the above functions

    Parameters
    ----------
    input_map_1 : either numpy.ndarray or path to mrc file
        First halfmap
    input_map_2 : either numpy.ndarray or path to mrc file
        Second halfmap

    Returns
    -------
    fsc_curve : numpy.ndarray

    '''
    import numpy as np
    from emmer.ndimage.fsc.fsc_utils import calculate_phase_correlation
    from emmer.ndimage.radial_profile.compute_radial_profile import compute_radial_profile
    from emmer.ndimage.map_utils import parse_input
    
    emmap_1 = parse_input(input_map_1)
    emmap_2 = parse_input(input_map_2)
    
    fft_1 = np.fft.rfftn(emmap_1)
    fft_2 = np.fft.rfftn(emmap_2)
    
    _, radii = compute_radial_profile(emmap_1, return_indices=True)
    
    map_shape = emmap_1.shape
    
    _, phase_correlation, _ = calculate_phase_correlation(fft_1, fft_2, radii, map_shape)
    phase_correlation = np.array(phase_correlation)
    return np.array(phase_correlation)



## Script to calculate the Real Space Cross Correlation (RSCC) between two maps, or any two ndarrays.


def compute_real_space_correlation(input_map_1,input_map_2):
    '''
    Function to calculate the Real Space Cross Correlation (RSCC) between two maps, or any two ndarrays. 
    
    RSCC is calculated by standardizing two arrays by subtracting their mean and dividing by their standard deviation

    Parameters
    ----------
    array1 : numpy.ndarray
        
    array2 : numpy.ndarray
        

    Returns
    -------
    RSCC : float
        Floating point number between 0 and 1 showing the RSCC between two arrays

    '''
    from emmer.ndimage.map_utils import parse_input
    array1 = parse_input(input_map_1)
    array2 = parse_input(input_map_2)
    
    (map1_mean,map1_std) = (array1.mean(),array1.std())
    (map2_mean,map2_std) = (array2.mean(),array2.std())
    
    n = array1.size
    
    RSCC = (((array1-map1_mean)*(array2-map2_mean))/(map1_std*map2_std)).sum() * (1/n)
    
    return RSCC


## Script to smoothen a binarised map using cosine edge falloff

def smoothen_mask(binarised_mask,cosine_falloff_length):
    '''
    binarised_mask: a binarised map
    cosine_falloff_length: length of cosine falloff
    '''
    from emmer.ndimage.filter.filter_utils import window3D
    from scipy import signal
    
    cosine_window_1d = signal.cosine(cosine_falloff_length)
    cosine_window_3d = window3D(cosine_window_1d)
    cosine_mask = signal.fftconvolve(binarised_mask,cosine_window_3d,mode='same')
    cosine_mask = cosine_mask/cosine_mask.max()
    return cosine_mask



import numpy as np

def estimate_bfactor_distribution(emmap_path, binarised_mask, wilson_cutoff, fsc_cutoff, boxsize=None, num_centers=1000, standard_notation=True):
    '''
    Script to estimate the local bfactor distribution of a cryo EM density map by random sampling of points in the map.

    emmap_path: path to the cryo EM density map
    binarised_mask: a binary mask of the same size as the cryo EM density map
    wilson_cutoff: the Wilson score cutoff for the local bfactor estimation
    fsc_cutoff: the FSC cutoff for the local bfactor estimation
    boxsize: the size of the box to extract around each point in the map
    num_centers: the number of points to extract from the map
    standard_notation: whether to use the standard notation for the bfactor (True) 

    returns:
    bfactors: the estimated bfactors (numpy array)
    
    '''

    from emmer.ndimage.radial_profile.estimate_bfactor_standard import estimate_bfactor_standard
    from emmer.ndimage.radial_profile.compute_radial_profile import compute_radial_profile
    from emmer.ndimage.radial_profile.frequency_array import frequency_array
    from emmer.ndimage.map_utils import  get_all_voxels_inside_mask, extract_window, load_map
    from emmer.include.emmer_utils import round_up_to_even
    import random
    from tqdm import tqdm
    
    emmap, apix = load_map(emmap_path)   
    
    if boxsize is None:
        boxsize = round_up_to_even(25 / apix)
    else:
        boxsize = round_up_to_even(boxsize)
        
    all_points = get_all_voxels_inside_mask(binarised_mask)
    random_centers = random.sample(all_points,num_centers)
    
    bfactor_distributions = {}
    
    for center in tqdm(random_centers, desc="Analysing local bfactors distribution"):
        try:            
            emmap_window = extract_window(emmap, center, boxsize)
            rp_local = compute_radial_profile(emmap_window)
            freq = frequency_array(rp_local, apix)                        
            bfactor,qfit = estimate_bfactor_standard(freq, rp_local, wilson_cutoff, fsc_cutoff, standard_notation=standard_notation, return_fit_quality=True)
            bfactor_distributions[tuple(center)] = tuple([bfactor, qfit])
        except Exception as e:
            print("Error at {}".format(center))
            print(e)
            raise
    
    return bfactor_distributions




import numpy as np
import mrcfile

def calculate_fourier_frequencies(im, apix):
    """Return the image frequency for every voxel in Fourierspace
    for n-dimensional images
    """
    per_axis_freq = [np.fft.fftfreq(N) for N in im.shape[:-1]]
    per_axis_freq.append(np.fft.rfftfreq(im.shape[-1]))
    XX, YY, ZZ = np.meshgrid(*per_axis_freq, indexing='ij', sparse=False)
    # fourier_frequencies = np.sqrt(np.sum([dim**2 for dim in dims]))
    fourier_frequencies = np.sqrt(XX**2 + YY**2 + ZZ**2)
    fourier_frequencies_angstrom = fourier_frequencies / apix
    return fourier_frequencies_angstrom


def tanh_filter(im_freq, cutoff):
    """Returns filter coefficients for a hyperbolic tangent filter. 
    """
    cutoff_freq = 1/cutoff
    filter_fall_off = 0.1;
    filter_coefficients = 1.0 - (1.0 - 0.5*(np.tanh((np.pi*(im_freq+cutoff_freq)/(2*filter_fall_off*cutoff_freq))) - np.tanh((np.pi*(im_freq-cutoff_freq)/(2*filter_fall_off*cutoff_freq)))));
    return filter_coefficients;

def get_fsc_filter(input_map_1, input_map_2):
    from emmer.ndimage.fsc.calculate_fsc_curve import calculate_fsc_curve
    import numpy as np
    fsc_curve = calculate_fsc_curve(input_map_1, input_map_2)
    C_ref = 2*fsc_curve / (1+fsc_curve)
    
    return C_ref

def window3D(w):
    # Convert a 1D filtering kernel to 3D
    # eg, window3D(numpy.hanning(5))
    L=w.shape[0]
    m1=np.outer(np.ravel(w), np.ravel(w))
    win1=np.tile(m1,np.hstack([L,1,1]))
    m2=np.outer(np.ravel(w),np.ones([1,L]))
    win2=np.tile(m2,np.hstack([L,1,1]))
    win2=np.transpose(win2,np.hstack([1,2,0]))
    win=np.multiply(win1,win2)
    return win



      



def compute_confidence_map(emmap, apix, window_size,fdr=0.01, lowPassFilter_resolution=None,remove_temp_files=True, return_fdr_value=False):
    '''
    Compute a confidence map from a EM map
    
    Parameters
    ----------
    emmap   : numpy.ndimage
        emmap data 
    apix : float
        Pixel size of the EM map
    window_size : int
        Window size for the confidence map
    apix : float
        Pixel size of the EM map
    fdr : float, optional
        False discovery rate. The default is 0.01.
    lowPassFilter_resolution : float, optional
        Low pass filter resolution. The default is None.
    remove_temp_files : bool, optional
        Remove temporary files. The default is True.
    
    Returns
    -------
    confidence_map : ndarray
        Confidence map
    fdr_threshold : float
        FDR threshold
    '''
    import numpy as np
    from emmer.include.confidenceMapUtil.confidenceMapMain import calculateConfidenceMap
    import os, shutil, time
    
    current_cwd = os.getcwd()
    timestamp =  str(time.time())
    temp_dir = current_cwd + '/fdr_output_temp_'+timestamp
    os.mkdir(temp_dir)
    os.chdir(temp_dir)
    confidenceMap,locFiltMap,locScaleMap,binMap,maskedMap = calculateConfidenceMap(
        em_map=emmap,apix=apix,noiseBox=None,testProc=None,ecdf=None,
        lowPassFilter_resolution=lowPassFilter_resolution,method=None, 
        window_size=window_size,windowSizeLocScale=None, locResMap=None,
        meanMap=None,varMap=None,fdr=fdr,modelMap=None,stepSize=None,mpi=None)
    
    fdr_threshold = np.min(maskedMap[np.nonzero(maskedMap)])
    
    os.chdir(current_cwd)
    if remove_temp_files:
        print("Clearing temporary files")
        shutil.rmtree(temp_dir)
    
    if return_fdr_value:
        return confidenceMap, fdr_threshold
    else:
        return confidenceMap


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 25 April 2022

@author: alok
"""

import unittest
import gemmi
import mrcfile
import numpy as np
import os

class test_fsc_tools(unittest.TestCase):
    def setUp(self):
        import pickle
        import emmer
        
        emmer_module_path = os.path.dirname(emmer.__file__)
        emmer_install_path = emmer_module_path
        self.emmer_install_path = emmer_install_path
        os.sys.path.insert(0, self.emmer_install_path)
        
        self.data_folder = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','data') 
        self.emmap_path = os.path.join(self.data_folder,'emd5778_map.mrc')
        self.mask_path = os.path.join(self.data_folder,'emd5778_mask.mrc')
        self.pdb_path = os.path.join(self.data_folder,'pdb3j5p.pdb')
        self.modmap_path = os.path.join(self.data_folder,'pdb3j5p_map.mrc')
        
        verify_output_file = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','verify','ndimage_outputs.pickle') 
        with open(verify_output_file, "rb") as outfile:
            self.validate = pickle.load(outfile)
        
    def copy_files(self, file_path, tempDir):
        from subprocess import run
        run(["cp",file_path,tempDir])
        if os.path.exists(os.path.join(tempDir, os.path.basename(file_path))):               
            return os.path.join(tempDir, os.path.basename(file_path))
            
        else:
            raise UserWarning("Could not copy {} to {}".format(path,tempDir))
        


    def test_compute_fsc(self):
        ## shake pdb within mask at 2A resolution
        from emmer.pdb.pdb_tools.perturb_within_mask import perturb_within_mask
        from emmer.ndimage.fsc.calculate_fsc_curve import calculate_fsc_curve
        from emmer.ndimage.mask.get_atomic_model_mask import get_atomic_model_mask
        from emmer.ndimage.map_utils import load_map, save_as_mrc
        from emmer.ndimage.radial_profile.frequency_array import frequency_array
        from emmer.ndimage.radial_profile.profile_tools import get_index_at_resolution
        from emmer.pdb.convert.convert_pdb_to_map import convert_pdb_to_map
        import matplotlib.pyplot as plt
        
        pdb_path = self.pdb_path
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_pdb_path = self.copy_files(pdb_path, tmpdir)
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)

            ## get atomic model mask
            emmap, apix = load_map(copied_sample_emmap_path)
            model_mask = get_atomic_model_mask(emmap=emmap, apix=apix,pdb_path=copied_pdb_path, dilation_radius=3, softening_parameter=5)
            save_as_mrc(model_mask, os.path.join(tmpdir,'model_mask.mrc'),apix=1.2156)
            model_mask_path = os.path.join(tmpdir,'model_mask.mrc')
            ## shake pdb within mask

            rmsd_mag = np.random.uniform(low=2, high=5)
            shaken_pdb_path_1 = perturb_within_mask(copied_pdb_path, model_mask_path, rmsd_magnitude=rmsd_mag)
            shaken_pdb_path_2 = perturb_within_mask(copied_pdb_path, model_mask_path, rmsd_magnitude=rmsd_mag)

            ## simulate model map
            model_map_1 = convert_pdb_to_map(input_pdb=shaken_pdb_path_1, apix=1.2156, size=(256,256,256))
            model_map_2 = convert_pdb_to_map(input_pdb=shaken_pdb_path_2, apix=1.2156, size=(256,256,256))

            ## calculate fsc maps
            fsc_maps = calculate_fsc_curve(model_map_1, model_map_2)
            freq = frequency_array(fsc_maps, apix=1.2156)
            
            fsc_index = get_index_at_resolution(freq, rmsd_mag)

            fsc_0 = fsc_maps[0]
            fsc_rmsd = fsc_maps[fsc_index]

            self.assertTrue(fsc_0 > fsc_rmsd)
            self.assertAlmostEqual(fsc_0, 1, delta=0.01)

    def test_apply_fsc_filter(self):
        ## shake pdb within mask at 2A resolution
        from emmer.pdb.pdb_tools.perturb_within_mask import perturb_within_mask
        from emmer.ndimage.fsc.calculate_fsc_curve import calculate_fsc_curve
        from emmer.ndimage.filter.apply_fsc_filter import apply_fsc_filter
        from emmer.ndimage.mask.get_atomic_model_mask import get_atomic_model_mask
        from emmer.ndimage.map_utils import load_map, save_as_mrc
        from emmer.ndimage.radial_profile.frequency_array import frequency_array
        from emmer.ndimage.radial_profile.profile_tools import get_index_at_resolution
        from emmer.ndimage.radial_profile.compute_radial_profile import compute_radial_profile
        from emmer.pdb.convert.convert_pdb_to_map import convert_pdb_to_map
        import matplotlib.pyplot as plt
        
        pdb_path = self.pdb_path
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_pdb_path = self.copy_files(pdb_path, tmpdir)
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap, apix = load_map(copied_sample_emmap_path)

            ## get atomic model mask
            model_mask = get_atomic_model_mask(emmap=emmap, apix=apix,pdb_path=copied_pdb_path, dilation_radius=3, softening_parameter=5)
            save_as_mrc(model_mask, os.path.join(tmpdir,'model_mask.mrc'),apix=1.2156)
            model_mask_path = os.path.join(tmpdir,'model_mask.mrc')
            ## shake pdb within mask

            rmsd_mag = np.random.uniform(low=2, high=5)
            shaken_pdb_path_1 = perturb_within_mask(copied_pdb_path, model_mask_path, rmsd_magnitude=rmsd_mag)
            shaken_pdb_path_2 = perturb_within_mask(copied_pdb_path, model_mask_path, rmsd_magnitude=rmsd_mag)

            ## simulate model map
            model_map_1 = convert_pdb_to_map(input_pdb=shaken_pdb_path_1, apix=1.2156, size=(256,256,256))
            model_map_2 = convert_pdb_to_map(input_pdb=shaken_pdb_path_2, apix=1.2156, size=(256,256,256))

            fsc_filtered_emmap = apply_fsc_filter(emmap, apix=1.2156, halfmap_1=model_map_1, halfmap_2=model_map_2)

            rp_emmap = compute_radial_profile(emmap)
            rp_fsc_filtered_emmap = compute_radial_profile(fsc_filtered_emmap)
            freq = frequency_array(rp_emmap, apix=1.2156)

            rmsd_index = get_index_at_resolution(freq, rmsd_mag)

            rp_emmap_0 = rp_emmap[0]
            rp_fsc_filtered_emmap_0 = rp_fsc_filtered_emmap[0]

            rp_emmap_rmsd = rp_emmap[rmsd_index]
            rp_fsc_filtered_emmap_rmsd = rp_fsc_filtered_emmap[rmsd_index]

            self.assertTrue(rp_emmap_rmsd > rp_fsc_filtered_emmap_rmsd)
            self.assertAlmostEqual(rp_emmap_0, rp_fsc_filtered_emmap_0, delta=0.001)
       





            


        
        



## Script to compute resolution from an amplitude

def resolution_from_amplitude(freq, amplitudes, probe_amplitude):
    from emmer.include.emmer_utils import probe_value_from_y
    xdata = freq
    ydata = amplitudes
    if probe_amplitude <= 0:
        raise UserWarning("Enter probe amplitude > 0")
    y = probe_amplitude
    x = probe_value_from_y(y, xdata, ydata)
    probe_resolution = 1/x

    return probe_resolution


import unittest
import os
def download_test_data_from_url(download_folder):
    import wget
    print("\nDownloading test data \n")
    url_test_data = "https://surfdrive.surf.nl/files/index.php/s/I36mEBPiu40nXa5/download"  ## SURFdrive URL    
    wget.download(url_test_data, download_folder, )

def download_verify_data_from_url(download_folder):
    import wget
    print("\nDownloading verification data \n")
    url_verify_data = "https://surfdrive.surf.nl/files/index.php/s/bbacbB4dTL64l0W/download"  ## SURFdrive URL
    wget.download(url_verify_data, download_folder)

def extract_tar_files_in_folder(tar_folder, use_same_folder=True):
    import tarfile
    import os
    if use_same_folder == 0:
        target_folder = tar_folder
    else:
        target_folder = os.path.dirname(tar_folder)

    for file in os.listdir(tar_folder):
        if file.endswith(".tar.gz"):
            print("Extracting: {}".format(file))
            tar = tarfile.open(os.path.join(tar_folder,file))
            tar.extractall(target_folder)
            tar.close()

def run_tests():
    import emmer
    import os
    ## Make sure the test data is downloaded
    emmer_path = os.path.dirname(os.path.abspath(emmer.__file__))
    test_data_folder = os.path.join(emmer_path, "tests", "ndimage_unittest","data")
    test_verify_data_folder = os.path.join(emmer_path, "tests", "ndimage_unittest","verify")

    unittest_folder = os.path.join(emmer_path, "tests")
    
    if not os.path.exists(test_data_folder):
        os.makedirs(test_data_folder, exist_ok=True)
        download_test_data_from_url(test_data_folder)
    if not os.path.exists(test_verify_data_folder):
        os.makedirs(test_verify_data_folder, exist_ok=True)
        download_verify_data_from_url(test_verify_data_folder)
    
    ## Extract the tar files
    extract_tar_files_in_folder(test_data_folder, use_same_folder=True)
    extract_tar_files_in_folder(test_verify_data_folder, use_same_folder=True)

    ## Create the test suite
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover(unittest_folder, pattern="test_*.py")
    
    ## Run the test suite
    test_runner = unittest.TextTestRunner(verbosity=0)
    test_result = test_runner.run(test_suite)
    print("test_result: {}".format(test_result))
    if test_result.wasSuccessful():
        print("All tests passed")
    else:
        print("Some tests failed")
        print(test_result.printErrors())
    
    print("="*80)
    print("Done")
    print("="*80)








## Script to generate different mask from an MRC file
import numpy as np
def get_atomic_model_mask(emmap, apix, pdb_path, dilation_radius=3, softening_parameter=5,  verbose=False):
    '''
    Function to generate a mask in the same grid as an emmap file based on atomic positions in a PDB file

    Parameters
    ----------
    emmap_path : str
        Path to a reference emmap for metadata
    pdb_path : str
        Path to a PDB/MMCIF file
  
    dilation_radius : float, optional
        The radius of the sphere (in Ang) to place at atomic positions determined by the PDB file. Default is 3A.

    Returns
    -------
    model_mask : str
    

    '''
    from emmer.ndimage.mask.mask_utils import get_atomic_point_map
    from emmer.ndimage.map_utils import dilate_mask
    from emmer.ndimage.filter.smoothen_mask import smoothen_mask
        
    map_shape = emmap.shape

    atomic_point_map = get_atomic_point_map(pdb_path, map_shape, apix)    
        
    dilation_radius_int = round(dilation_radius / apix)
    dilated_mask = dilate_mask(atomic_point_map, radius=dilation_radius_int)
    
    softened_mask = smoothen_mask(dilated_mask, softening_parameter)

    return softened_mask


## Script to apply a radial profile to a given EM density map

def set_radial_profile_to_volume(emmap, ref_profile):
    '''
    emmap : numpy.ndarray (dims=3)
        Input map
    ref_profile : numpy.ndarray (dims=1)
        Reference profile
    
    Returns
    -------
    scaled_map : numpy.ndarray (dims=3)
    
        '''
    from emmer.ndimage.sharpen.scaling_tools import compute_radial_profile_simple, compute_scale_factors, set_radial_profile_simple
    em_profile,frequencies = compute_radial_profile_simple(emmap, return_frequencies=True)
    scale_factors = compute_scale_factors(em_profile, ref_profile)
    scaled_map = set_radial_profile_simple(emmap, scale_factors, frequencies)
    
    return scaled_map


## Script to measure the debye effect magnitude from a radial profile

def measure_debye_effect_magnitude(freq, amplitudes, wilson_cutoff, fsc_cutoff, wilson_range=None, return_exponential_fit=False):
    '''
    Function to measure the debye effect magnitude from a radial profile
    freq : array
        Frequency axis
    amplitudes : array
        Amplitude axis
    wilson_cutoff : float   
        Wilson cutoff
    fsc_cutoff : float
        FSC cutoff
    wilson_range : array
        Range of frequencies to use for the wilson region
    return_exponential_fit : bool
        If true, return the exponential fit
    
    '''
    from emmer.ndimage.radial_profile.profile_tools import measure_wilson_deviation
    wilson_deviations = measure_wilson_deviation(freq, amplitudes, wilson_cutoff, fsc_cutoff, wilson_range, return_exponential_fit)
    absolute_deviations = abs(wilson_deviations)
    return absolute_deviations.sum()
    


## Script to count the number of segments in a map at a given threshold

def count_segments(emmap, reference_threshold):
    from skimage import measure
    import numpy as np

    binarised_emmap = (emmap>reference_threshold).astype(np.int_)
    labels, num_regions = measure.label(binarised_emmap, background=0, return_num=True)
    
    return num_regions
    


## Script to scale a target profile to match B factor of a reference profile

def scale_profiles(reference_profile_tuple, target_profile_tuple, wilson_cutoff, fsc_cutoff, return_bfactor_properties=False):
    '''
    Function to scale an input theoretical profile to a reference profile

    Parameters
    ----------
    reference_profile_tuple : tuple
        (freq_reference, amplitude_reference)
    target_profile_tuple : tuple
        (freq_theoretical, amplitude_theoretical)
    just_use_exponential : bool, optional
        Returns just an exponential fit and not a scaled profile
    using_reference_profile : TYPE, optional
        DESCRIPTION. The default is False.
    start_freq : TYPE, optional
        DESCRIPTION. The default is 0.3.

    Returns
    -------
    TYPE
        DESCRIPTION.

    '''
    import numpy as np
    from emmer.ndimage.radial_profile.estimate_bfactor_standard import estimate_bfactor_standard
    
    freq = reference_profile_tuple[0]
    reference_amplitude = reference_profile_tuple[1]
    
    freq_scale = target_profile_tuple[0]
    scale_amplitude = target_profile_tuple[1]
 
    bfactor_reference, fit_amp_reference, quality_of_fit = estimate_bfactor_standard(freq, reference_amplitude, wilson_cutoff=wilson_cutoff, fsc_cutoff=fsc_cutoff, return_amplitude=True,return_fit_quality=True)
    bfactor_scale, fit_amp_scale = estimate_bfactor_standard(freq_scale, scale_amplitude, wilson_cutoff=wilson_cutoff, fsc_cutoff=fsc_cutoff, return_amplitude=True)
    
    bfactor_diff = bfactor_reference-bfactor_scale
    
    amp_scaling_factor = fit_amp_reference / fit_amp_scale
        
    amplitude_scaled = amp_scaling_factor * scale_amplitude * np.exp(bfactor_diff * freq**2 / 4)
    
    
    if return_bfactor_properties:
        return (freq,amplitude_scaled), (bfactor_reference, fit_amp_reference, quality_of_fit)
    else:
        return (freq, amplitude_scaled)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 25 April 2022

@author: alok
"""

import unittest
import gemmi
import mrcfile
import numpy as np
import os

class test_contour_tools(unittest.TestCase):
    def setUp(self):
        import pickle
        import emmer
        
        emmer_module_path = os.path.dirname(emmer.__file__)
        emmer_install_path = emmer_module_path
        self.emmer_install_path = emmer_install_path
        os.sys.path.insert(0, self.emmer_install_path)
        
        self.data_folder = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','data') 
        self.emmap_path = os.path.join(self.data_folder,'emd5778_map.mrc')
        self.mask_path = os.path.join(self.data_folder,'emd5778_mask.mrc')
        self.pdb_path = os.path.join(self.data_folder,'pdb3j5p.pdb')
        self.modmap_path = os.path.join(self.data_folder,'pdb3j5p_map.mrc')
        
        verify_output_file = os.path.join(self.emmer_install_path, 'tests','ndimage_unittest','verify','ndimage_outputs.pickle') 
        with open(verify_output_file, "rb") as outfile:
            self.validate = pickle.load(outfile)
        
    def copy_files(self, file_path, tempDir):
        from subprocess import run
        run(["cp",file_path,tempDir])
        if os.path.exists(os.path.join(tempDir, os.path.basename(file_path))):               
            return os.path.join(tempDir, os.path.basename(file_path))
            
        else:
            raise UserWarning("Could not copy {} to {}".format(path,tempDir))
        


    def test_surface_area_threshold(self):
        ## shake pdb within mask at 2A resolution
        from emmer.ndimage.contour.surface_area_at_threshold import surface_area_at_threshold
        from emmer.ndimage.map_utils import load_map
        
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap,_ = load_map(copied_sample_emmap_path)
            ref_threshold_1 = 0.005
            ref_threshold_2 = 0.01

            ref_area_1 = 134.4e3 ## from chimeraX  in A^2
            ref_area_2 = 72.44e3  ## from chimeraX in A^2

            ## get surface area threshold
            surface_area_1 = surface_area_at_threshold(emmap, apix=1.2156, reference_threshold=ref_threshold_1)
            surface_area_2 = surface_area_at_threshold(emmap, apix=1.2156, reference_threshold=ref_threshold_2)

            surface_area_relative_difference_1 = abs(surface_area_1 - ref_area_1)/ref_area_1
            surface_area_relative_difference_2 = abs(surface_area_2 - ref_area_2)/ref_area_2

            self.assertLess(surface_area_relative_difference_1, 0.15)
            self.assertLess(surface_area_relative_difference_2, 0.15) 
    
    def test_count_number_of_segments(self):
        from emmer.ndimage.contour.count_segments import count_segments
        from emmer.ndimage.map_utils import load_map
        
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap,_ = load_map(copied_sample_emmap_path)
            ref_threshold = 0.01
            ref_num_regions = 5

            ## get number of segments
            num_regions = count_segments(emmap, reference_threshold=ref_threshold)

            self.assertEqual(num_regions, ref_num_regions)
    
    def test_volume_at_threshold(self):
        from emmer.ndimage.contour.volume_at_threshold import volume_at_threshold
        from emmer.ndimage.map_utils import load_map
        
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap,apix = load_map(copied_sample_emmap_path)
            ref_threshold = 0.01
            ref_volume = 96.5e3 ## from chimeraX in A^3 (approx)

            ## get volume at threshold
            volume = volume_at_threshold(emmap, apix, reference_threshold=ref_threshold)

            error = abs(volume - ref_volume)/ref_volume

            self.assertLess(error, 0.05)
    
    def test_volume_matching_threshold(self):
        from emmer.ndimage.contour.volume_matching_threshold import volume_matching_threshold
        from emmer.ndimage.map_utils import load_map
        
        sample_emmap_path = self.emmap_path

        ## create temporary directory
        from tempfile import TemporaryDirectory

        with TemporaryDirectory() as tmpdir:
            copied_sample_emmap_path = self.copy_files(sample_emmap_path, tmpdir)
            emmap,apix = load_map(copied_sample_emmap_path)
            ref_threshold = 0.01210
            ref_volume = 70e3

            ## get volume at threshold
            threshold = volume_matching_threshold(emmap, apix, reference_volume=ref_volume, min_threshold=0.01, max_threshold=0.03, num_bins=20)

            self.assertAlmostEqual(threshold, ref_threshold, delta=0.01)





    
        

            
    

       

       











            




## Script to apply FSC filter to EMMAP based on either FSC curve or two halfmaps

def apply_fsc_filter(emmap, apix, fsc_curve=None, halfmap_1=None, halfmap_2=None):    
    '''
    Function to apply a FSC filter to a cryo EM map from either a FSC curve or two halfmaps.
    emmap : numpy.ndarray
        Cryo EM map
    apix : float
        Pixel size in Angstroms
    fsc_curve : numpy.ndarray
        FSC curve
    halfmap_1 : numpy.ndarray or path to halfmap 1
        First halfmap
    halfmap_2 : numpy.ndarray or path to halfmap 2
        Second halfmap

    '''
    from emmer.ndimage.sharpen.scaling_tools import compute_radial_profile_simple
    from emmer.ndimage.sharpen.set_radial_profile_to_volume import set_radial_profile_to_volume
    from emmer.ndimage.filter.filter_utils import get_fsc_filter
    import numpy as np
    
    if fsc_curve is not None:
        C_ref = 2*fsc_curve / (1+fsc_curve)
    elif halfmap_1 is not None and halfmap_2 is not None:
        C_ref = get_fsc_filter(halfmap_1, halfmap_2)
    
    
    rp_emmap = compute_radial_profile_simple(emmap, return_frequencies=False)
    fsc_filtered_rp = rp_emmap * C_ref
    
    filtered_emmap = set_radial_profile_to_volume(emmap, fsc_filtered_rp)

    return filtered_emmap


## Script to perform map sharpening using B factors

def sharpen_maps(emmap, apix, sharpening_bfactor):
    '''
    Function to apply a global sharpening factor to EM density maps 

    Parameters
    ----------
    emmap : numpy.ndarray (dims=3)
        Input map
    apix : Float
        Pixelsize (one dimension only)
    sharpening_bfactor : float
        Global sharpening factor (negative if you want to sharpen, positive if you want to blur)

    Returns
    -------
    sharpened_map : numpy.ndarray (dims=3)

    '''
    import numpy as np
    from emmer.ndimage.radial_profile.frequency_array import frequency_array
    from emmer.ndimage.sharpen.set_radial_profile_to_volume import set_radial_profile_to_volume
    from emmer.ndimage.sharpen.scaling_tools import compute_radial_profile_simple
        
    emmap_profile = compute_radial_profile_simple(emmap)
    freq = frequency_array(amplitudes=emmap_profile, apix=apix)
    
    sharpened_profile = emmap_profile * np.exp(-1*sharpening_bfactor/4 * freq**2)

    sharpened_map = set_radial_profile_to_volume(emmap, sharpened_profile)
    
    return sharpened_map


# -*- coding: utf-8 -*-
"""
Created on Sun Jul 18 14:58:49 2021

"""

# convert_to_mmcif contains a function to convert .pdb file format to .mmcif
# file format using gemmi, as well as functions for the saving of a PDB 
# structure in either format

# global imports
import gemmi

#%% functions
def convert_pdb_to_mmcif(pdb_path, mmcif_path=None):
    """Convert coordinate file from PDB format to mmCIF/PDBx format"""
    # local imports
    import os
    
    structure = gemmi.read_structure(pdb_path)
    if mmcif_path is None:
        mmcif_path = os.path.splitext(pdb_path)[0] + '.cif'
    write_structure_as_mmcif(structure, mmcif_path)
    return mmcif_path


def write_structure_as_mmcif(structure, mmcif_name):
    """Write a Gemmi structure out to an mmCIF file."""
    structure.make_mmcif_document().write_file(mmcif_name)


def write_structure_as_pdb(structure, pdb_name):
    """Write a Gemmi structure out to a PDB file."""
    structure.write_pdb(pdb_name)






import numpy as np
import math


def round_up_to_even(x):
    ceil_x = math.ceil(x)
    if ceil_x % 2 == 0:   ## check if it's even, if not return one higher
        return ceil_x
    else:
        return ceil_x+1

def round_up_to_odd(x):
    ceil_x = math.ceil(x)
    if ceil_x % 2 == 0:   ## check if it's even, if so return one higher
        return ceil_x+1
    else:
        return ceil_x

def true_percent_probability(n):
    x = np.random.uniform(low=0, high=100)
    if x <= n:
        return True
    else:
        return False

def copy_file_to_folder(full_path_to_file, new_folder):
    import shutil
    import os
    
    source = full_path_to_file
    file_name = os.path.basename(source)
    destination = os.path.join(new_folder, file_name)
    shutil.copyfile(source, destination)
    
    return destination

def linear(x,a,b):
    return a * x + b

def general_quadratic(x,a,b,c):
    return a * x**2 + b*x + c
    

def round_up_proper(x):
    epsilon = 1e-15  ## To round up in case of rounding to odd
    return np.round(x+epsilon).astype(int)


## Function to probe the value of Y from a given X value
def probe_value_from_x(x, xdata, ydata):
    from scipy.interpolate import interp1d
    ## Make sure x is between the min and max of xdata
    x = np.clip(x, xdata.min(), xdata.max())

    f = interp1d(xdata, ydata)
    y_probe = f(x)
    return y_probe

## Function to probe the value of X from a given Y value
def probe_value_from_y(y, xdata, ydata):
    from scipy.interpolate import interp1d
    ## Make sure y is between the min and max of ydata
    y = np.clip(y, ydata.min(), ydata.max())

    f = interp1d(ydata, xdata)
    x_probe = f(y)
    return x_probe



## Script to compute a shaken PDB structure within a molecular boundary


def perturb_within_mask(pdb_path, mask_path, rmsd_magnitude,mask_threshold=0.5, verbose=False):
    '''
    Function to perturb a PDB structure within a molecular boundary
    
    Parameters
    ----------
    pdb_path : str
        Path to a PDB/MMCIF file
    mask_path : str
        Path to a MRC file
    rmsd_magnitude : float
        The magnitude of the RMSD to apply to the PDB structure
    mask_threshold : float, optional
        The threshold for the mask. The default is 0.5.
    verbose : bool, optional
        Verbose. The default is False.
    
    Returns
    -------
    shaken_pdb_path : str
        Path to the shaken PDB structure
    '''
    from emmer.pdb.pdb_tools.shake_pdb_within_mask import shake_pdb_within_mask

    shaken_pdb_path = shake_pdb_within_mask(pdb_path, mask_path, rmsd_magnitude, mask_threshold=mask_threshold, verbose=verbose)
    return shaken_pdb_path


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 29 10:16:11 2021

"""

# global imports
import gemmi
from emmer.pdb.pdb_tools import add_cryst1_line

def add_cryst1_line(pdb_path,unitcell=None,emmap_path=None,new_pdb_path=None):
    '''
    pdb_path -> Address of .pdb path
    
    Some PDB files developed for cryoEM maps do not have proper cryst1 record. Two options to modify:

    1. From an input tuple, or array. In this case, unitcell is a python tuple, which has unit cell dimensions in angstorm
    Ex: unitcell = (x,y,z)
    2. From a mrcfile. In this case, point to an associated EM map and the unit cell dimensions are taken from that
    emmap_path -> Address of associated .mrc file
    
    If you like to the pdb file with a different name, or address then change the 'new_pdb_path' 
    
    '''
    if emmap_path is not None:
        mrc = mrcfile.open(emmap_path)
        cella = mrc.header.cella
        x = cella.x
        y = cella.y
        z = cella.z
    elif unitcell is not None:
        x = unitcell[0]
        y = unitcell[1]
        z = unitcell[2]
    else:
        print("Please give either unit cell dimensions (in Ang) or point to an associated mrc file!")
        return
    
    unitcell = gemmi.UnitCell(x,y,z,90,90,90)
    
    gemmi_structure = gemmi.read_structure(pdb_path)
    gemmi_structure.cell = unitcell
    if new_pdb_path is None:
        gemmi_structure.write_pdb(pdb_path)
    else:
        gemmi_structure.write_pdb(new_pdb_path)
           






def get_atomic_point_map(input_pdb, map_shape, apix):
    import numpy as np

    from emmer.pdb.convert.convert_pdb_to_mrc_position import convert_pdb_to_mrc_position
    from emmer.pdb.pdb_utils import detect_pdb_input, get_all_atomic_positions
    
    structure = detect_pdb_input(input_pdb)
    mask = np.zeros(map_shape)
    pdb_positions = get_all_atomic_positions(structure)        
    mrc_position = convert_pdb_to_mrc_position(pdb_positions, apix)
    zz,yy,xx = zip(*mrc_position)
    mask[zz,yy,xx] = 1
    
    return mask




import os
import sys
import argparse
from argparse import RawTextHelpFormatter
from datetime import datetime
import pyfiglet

progname = os.path.basename(sys.argv[0])
author = 'Alok Bharadwaj, Maarten Joosten, Stefan T Huber, Arjen Jakobi, Reinier de Bruin'
version = "0.1"

test_run_command = "emmer test"

description = "A python toolkit for the cryo-EM developer"

main_parser = argparse.ArgumentParser(prog="emmer", description=description, formatter_class=RawTextHelpFormatter)

##################### ADD TEST RUN COMMAND ##############################
main_parser.add_argument('--test', action='store_true', help='Run the test suite')

def print_start_banner():
    """
    Prints a start banner
    """
    from textwrap import fill
    import time
    try:
        username = os.environ['USER']
    except KeyError:
        username = 'User: unknown'
    
    today_date = datetime.now().strftime("%Y-%m-%d")
    time_now = datetime.now().strftime("%H:%M:%S")

    print("="*80)
    print("="*80)
    print_text = pyfiglet.figlet_format("EMmer", font="big")
    print(print_text)
    print("\t"*6 + "Version: v{}".format(version))
    print("."*80)
    print("  |  ".join(["User: {}".format(username), "Date: {}".format(today_date), "Time: {}".format(time_now)]))
    print("\n")
    print("Authors:\n")
    author_list = ["Arjen J. Jakobi (TU Delft)",  "Alok Bharadwaj (TU Delft)", "Reinier de Bruin (TU Delft)", \
                "Maarten Joosten (TU Delft)", "Stefan T. Huber (TU Delft)"]
    for author in author_list:
        print("\t{} \n".format(author))
    print("Running tests. Please wait...")
    ## Delay for a bit to make sure the banner is printed before the tests start
    
    time.sleep(1)
    print("="*80)
    print("="*80)

    
def main():
    """
    Main function
    """
    from coverage import Coverage  ## To measure code coverage
    from emmer.tests.run import run_tests
    print_start_banner()
    cov = Coverage()
    cov.start()
    ############################## RUN TESTS ##############################
    run_tests()
    ############################## END TESTS ##############################
    cov.stop()
    cov.save()
    cov.html_report("htmlcov")


if __name__ == '__main__':
    main()



import numpy as np
from emmer.include.confidenceMapUtil import FDRutil
import argparse, math, os, sys
from argparse import RawTextHelpFormatter
import time
  
def compute_padding_average(vol, mask):
    mask = (mask > 0.5).astype(np.int8)
    average_padding_intensity = np.mean(np.ma.masked_array(vol, mask))
    return average_padding_intensity

def pad_or_crop_volume(vol, dim_pad=None, pad_value = None, crop_volume=False):   
    if (dim_pad == None):
        return vol
    else:
        dim_pad = np.round(np.array(dim_pad)).astype('int')

        if pad_value == None:
            pad_value = 0

        if (dim_pad[0] <= vol.shape[0] or dim_pad[1] <= vol.shape[1] or dim_pad[2] <= vol.shape[2]): 
            crop_volume = True
        
        if crop_volume:
            crop_vol = vol[int(vol.shape[0]/2-dim_pad[0]/2):int(vol.shape[0]/2+dim_pad[0]/2+dim_pad[0]%2), :, :]
            crop_vol = crop_vol[:, int(vol.shape[1]/2-dim_pad[1]/2):int(vol.shape[1]/2+dim_pad[1]/2+dim_pad[1]%2), :]
            crop_vol = crop_vol[:, :, int(vol.shape[2]/2-dim_pad[2]/2):int(vol.shape[2]/2+dim_pad[2]/2+dim_pad[2]%2)]
            
            return crop_vol
            
        else:
            pad_vol = np.pad(vol, ((int(dim_pad[0]/2-vol.shape[0]/2), int(dim_pad[0]/2-vol.shape[0]/2+dim_pad[0]%2)), (0,0), (0,0) ), 'constant', constant_values=(pad_value,))
            pad_vol = np.pad(pad_vol, ((0,0), (int(dim_pad[1]/2-vol.shape[1]/2), int(dim_pad[1]/2-vol.shape[1]/2+dim_pad[1]%2) ), (0,0)), 'constant', constant_values=(pad_value,))
            pad_vol = np.pad(pad_vol, ((0,0), (0,0), (int(dim_pad[2]/2-vol.shape[2]/2), int(dim_pad[2]/2-vol.shape[2]/2+dim_pad[2]%2))), 'constant', constant_values=(pad_value,))
            
            return pad_vol

def check_for_window_bleeding(mask, wn):
    masked_xyz_locs, masked_indices, mask_shape = get_xyz_locs_and_indices_after_edge_cropping_and_masking(mask, 0)
    
    zs, ys, xs = masked_xyz_locs.T
    nk, nj, ni = mask_shape

    if xs.min() < wn / 2 or xs.max() > (ni - wn / 2) or \
    ys.min() < wn / 2 or ys.max() > (nj - wn / 2) or \
    zs.min() < wn / 2 or zs.max() > (nk - wn / 2):
        window_bleed = True
    else:
        window_bleed = False
    
    return window_bleed

def get_xyz_locs_and_indices_after_edge_cropping_and_masking(mask, wn):

    mask = np.copy(mask);
    nk, nj, ni = mask.shape;

    kk, jj, ii = np.indices((mask.shape));
    kk_flat = kk.ravel();
    jj_flat = jj.ravel();
    ii_flat = ii.ravel();
    
    mask_bin = np.array(mask.ravel(), dtype=np.bool);
    indices = np.arange(mask.size);
    masked_indices = indices[mask_bin];
    cropped_indices = indices[(wn / 2 <= kk_flat) & (kk_flat < (nk - wn / 2)) &
                              (wn / 2 <= jj_flat) & (jj_flat < (nj - wn / 2)) &
                              (wn / 2 <= ii_flat) & (ii_flat < (ni - wn / 2))];
                                     
    cropp_n_mask_ind = np.intersect1d(masked_indices, cropped_indices);
    
    xyz_locs = np.column_stack((kk_flat[cropp_n_mask_ind], jj_flat[cropp_n_mask_ind], ii_flat[cropp_n_mask_ind]));

    return xyz_locs, cropp_n_mask_ind, mask.shape;

def prepare_mask_and_maps_for_scaling(emmap, modelmap, apix, wn_locscale, windowSize, method, locResMap, noiseBox):
    
    mask = np.zeros(emmap.shape);
        
    if mask.shape[0] == mask.shape[1] and mask.shape[0] == mask.shape[2] and mask.shape[1] == mask.shape[2]:
        rad = (mask.shape[0] // 2) ;
        z,y,x = np.ogrid[-rad: rad+1, -rad: rad+1, -rad: rad+1];
        mask = (x**2+y**2+z**2 <= rad**2).astype(np.int_).astype(np.int8);
        mask = pad_or_crop_volume(mask,emmap.shape);
        mask = (mask > 0.5).astype(np.int8);
    else:
        mask += 1;
        mask = mask[0:mask.shape[0]-1, 0:mask.shape[1]-1, 0:mask.shape[2]-1];
        mask = pad_or_crop_volume(emmap, (emmap.shape), pad_value=0);
           
    if wn_locscale is None:
        wn_locscale = int(round(7 * 3 * apix)); # set default window size to 7 times average resolution
    elif wn_locscale is not None:
        wn_locscale = int(math.ceil(wn_locscale / 2.) * 2);
   
    #wn = wn_locscale;
    if windowSize is None:
        wn = wn_locscale;
    elif windowSize is not None: 
        wn = int(math.ceil(windowSize / 2.) * 2);
	
    if method is not None:
        method = method;
    else:
        method = 'BY';
	
    if noiseBox is not None:
        boxCoord = noiseBox;
    else:
        boxCoord = 0;

    window_bleed_and_pad = check_for_window_bleeding(mask, wn_locscale);
    if window_bleed_and_pad:
        pad_int_emmap = compute_padding_average(emmap, mask);
        pad_int_modmap = compute_padding_average(modelmap, mask);
        map_shape = [(emmap.shape[0] + wn_locscale), (emmap.shape[1] + wn_locscale), (emmap.shape[2] + wn_locscale)];
        emmap = pad_or_crop_volume(emmap, map_shape, pad_int_emmap);
        modelmap = pad_or_crop_volume(modelmap, map_shape, pad_int_modmap);
        mask = pad_or_crop_volume(mask, map_shape, 0);
        if locResMap is not None:
            locResMap = pad_or_crop_volume(locResMap, map_shape, 100.0);
    
    #if wished so, do local filtration                      
    if locResMap is not None:
        locResMap[locResMap == 0.0] = 100.0;
        locResMap[locResMap >= 100.0] = 100.0;
        locFilt = True;    
    else:
        locFilt = False;
        locResMap = np.ones(emmap.shape);

    return emmap, modelmap, mask, wn, wn_locscale, window_bleed_and_pad, method, locFilt, locResMap, boxCoord;

def compute_radial_profile(volFFT, frequencyMap):

    dim = volFFT.shape;
    ps = np.real(np.abs(volFFT));
    frequencies = np.fft.rfftfreq(dim[0]);
    #frequencies = np.linspace(0, 0.5, int(math.ceil(dim[0]/2.0)));
    bins = np.digitize(frequencyMap, frequencies);
    bins = bins - 1;
    radial_profile = np.bincount(bins.ravel(), ps.ravel()) / np.bincount(bins.ravel())

    return radial_profile, frequencies;

def compute_scale_factors(em_profile, ref_profile):

    np.seterr(divide='ignore', invalid='ignore'); #no error for division by zero 
    #scale_factor = (ref_profile**2/em_profile**2);
    #scale_factor[ ~ np.isfinite( scale_factor )] = 0 #handle division by zero
    #scale_factor = np.sqrt(scale_factor);
    scale_factor = np.divide(np.abs(ref_profile), np.abs(em_profile));    
    scale_factor[ ~ np.isfinite( scale_factor )] = 0; #handle division by zero

    return scale_factor;

def set_radial_profile(volFFT, scaleFactors, frequencies, frequencyMap, shape): 

    scalingMap = np.interp(frequencyMap, frequencies, scaleFactors);	
    scaledMapFFT = scalingMap * volFFT;
    scaledMap = np.real(np.fft.irfftn(scaledMapFFT, shape, norm='ortho'));
    
    return scaledMap, scaledMapFFT;

def calculate_scaled_map(emmap, modmap, mask, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdfBool, stepSize):

    sizeMap = emmap.shape
    sharpened_map = np.zeros(sizeMap);
    sharpened_mean_vals = np.zeros(sizeMap);
    sharpened_var_vals = np.zeros(sizeMap);
    sharpened_ecdf_vals = np.zeros(sizeMap);
    central_pix = int(round(wn_locscale / 2.0));
    center = np.array([0.5*sizeMap[0], 0.5*sizeMap[1], 0.5*sizeMap[2]]);
    
    #get the background noise sample
    if boxCoord == 0:
        noiseMap = emmap[int(center[0]-0.5*wn):(int(center[0]-0.5*wn) + wn), int(0.02*wn+wn_locscale/2.0):(int(0.02*wn+wn_locscale/2.0) + wn), (int(center[2]-0.5*wn)):(int((center[2]-0.5*wn) + wn))];
    else:
        noiseMap = emmap[int(boxCoord[0]-0.5*wn +wn_locscale/2.0):(int(boxCoord[0]-0.5*wn + wn_locscale/2.0) + wn), int(boxCoord[1]-0.5*wn+ wn_locscale/2.0):(int(boxCoord[1]-0.5*wn + wn_locscale/2.0) + wn), (int(boxCoord[2]-0.5*wn + wn_locscale/2.0)):(int((boxCoord[2]-0.5*wn + wn_locscale/2.0)+wn))];


    #prepare noise map for scaling
    frequencyMap_noise = FDRutil.calculate_frequency_map(noiseMap);
    noiseMapFFT = np.fft.rfftn(noiseMap, norm='ortho');
    noise_profile, frequencies_noise = compute_radial_profile(noiseMapFFT, frequencyMap_noise);
    
    #prepare windows of particle for scaling
    frequencyMap_mapWindow = FDRutil.calculate_frequency_map(np.zeros((wn_locscale, wn_locscale, wn_locscale)));

    numSteps = len(range(0, sizeMap[0] - int(wn_locscale), stepSize))*len(range(0, sizeMap[1] - int(wn_locscale), stepSize))*len(range(0, sizeMap[2] - int(wn_locscale), stepSize));
    print("Sart LocScale. This might take a minute ...");
    counterSteps = 0;
    for k in range(0, sizeMap[0] - int(wn_locscale), stepSize):
        for j in range(0, sizeMap[1] - int(wn_locscale), stepSize):
            for i in range(0, sizeMap[2] - int(wn_locscale), stepSize):
                
                #print progress
                counterSteps = counterSteps + 1;
                progress = counterSteps/float(numSteps);                
                if counterSteps%(int(numSteps/20.0)) == 0:
                        output = "%.1f" %(progress*100) + "% finished ..." ;
                        print(output);                   

                #crop windows
                emmap_wn = emmap[k: k + wn_locscale, j: j + wn_locscale, i: i + wn_locscale];
                modmap_wn = modmap[k: k + wn_locscale, j: j + wn_locscale, i: i + wn_locscale];                


                #do sharpening of the sliding window
                emmap_wn_FFT = np.fft.rfftn(np.copy(emmap_wn), norm='ortho');
                modmap_wn_FFT = np.fft.rfftn(np.copy(modmap_wn), norm='ortho');
                em_profile, frequencies_map = compute_radial_profile(emmap_wn_FFT, frequencyMap_mapWindow);
                mod_profile, _ = compute_radial_profile(modmap_wn_FFT, frequencyMap_mapWindow);
                scale_factors = compute_scale_factors(em_profile, mod_profile);
                map_b_sharpened, map_b_sharpened_FFT = set_radial_profile(emmap_wn_FFT, scale_factors, frequencies_map, frequencyMap_mapWindow, emmap_wn.shape);
                  

                #scale noise window with the interpolated scaling factors
                mapNoise_sharpened, mapNoise_sharpened_FFT = set_radial_profile(np.copy(noiseMapFFT), scale_factors, frequencies_map, frequencyMap_noise, noiseMap.shape);

                #local filtering routines
                if locFilt == True:
                    tmpRes = round(apix/locResMap[k, j, i], 3);
                                                
                    mapNoise_sharpened = FDRutil.lowPassFilter(mapNoise_sharpened_FFT, frequencyMap_noise, tmpRes, noiseMap.shape);
                    map_b_sharpened = FDRutil.lowPassFilter(map_b_sharpened_FFT, frequencyMap_mapWindow, tmpRes, emmap_wn.shape);
                    
                    #calculate noise statistics	  
                    map_noise_sharpened_data = mapNoise_sharpened;	   
                        
                    if ecdfBool:
                        tmpECDF, sampleSort = FDRutil.estimateECDFFromMap(map_noise_sharpened_data, -1, -1);
                        ecdf = np.interp(map_b_sharpened[central_pix, central_pix, central_pix], sampleSort, tmpECDF, left=0.0, right=1.0); 
                    else:
                        ecdf = 0;

                    mean = np.mean(map_noise_sharpened_data);
                    var = np.var(map_noise_sharpened_data);

                    if var < 0.5:
                        var = 0.5;
                        mean = 0.0;
                    if tmpRes == round(apix/100.0, 3):  
                        mean = 0.0;
                        var = 0.0;
                        ecdf = 0;   	 			
                else:    
                    #calculate noise statistics
                    map_noise_sharpened_data = np.copy(mapNoise_sharpened);
                 
                    if ecdfBool:
                        tmpECDF, sampleSort = FDRutil.estimateECDFFromMap(map_noise_sharpened_data, -1, -1);
                        ecdf = np.interp(map_b_sharpened, sampleSort, tmpECDF, left=0.0, right=1.0); 
                    else:
                        ecdf = 0;
                                        
                    mean = np.mean(map_noise_sharpened_data);
                    var = np.var(map_noise_sharpened_data);
                    if var < 0.5:
                        var = 0.5;
                        mean = 0.0;

                #put values back into the the original maps
                halfStep=int((wn_locscale/2.0) - (stepSize/2.0));
                sharpened_map[k + halfStep : k + halfStep + stepSize, j + halfStep : j + halfStep + stepSize, i + halfStep : i + halfStep + stepSize] = np.copy(map_b_sharpened[halfStep:halfStep+stepSize, halfStep:halfStep+stepSize, halfStep:halfStep+stepSize]);
                sharpened_mean_vals[k + halfStep : k + halfStep + stepSize, j + halfStep : j + halfStep + stepSize, i + halfStep : i + halfStep + stepSize] = mean;
                sharpened_var_vals[k + halfStep :  k + halfStep + stepSize, j + halfStep : j + halfStep + stepSize, i + halfStep : i + halfStep + stepSize] = var;
                if ecdfBool:
                    sharpened_ecdf_vals[k + halfStep : k + halfStep + stepSize, j + halfStep : j + halfStep + stepSize, i + halfStep : i + halfStep + stepSize] = ecdf[halfStep:halfStep+stepSize, halfStep:halfStep+stepSize, halfStep:halfStep+stepSize];
                else:
                    sharpened_ecdf_vals[k + halfStep: k + halfStep + stepSize, j + halfStep: j + halfStep + stepSize,
                    i + halfStep: i + halfStep + stepSize] = 0.0;

    return sharpened_map, sharpened_mean_vals, sharpened_var_vals, sharpened_ecdf_vals;

def get_central_scaled_pixel_vals_after_scaling(emmap, modmap, masked_xyz_locs, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdfBool):

    sharpened_vals = [];
    sharpened_mean_vals = [];
    sharpened_var_vals = [];
    sharpened_ecdf_vals = []; 
       
    central_pix = int(round(wn_locscale / 2.0));
    sizeMap = emmap.shape;
    center = np.array([0.5*sizeMap[0], 0.5*sizeMap[1], 0.5*sizeMap[2]]);
    
    #get the background noise sample
    if boxCoord == 0:
        noiseMap = emmap[int(center[0]-0.5*wn):(int(center[0]-0.5*wn) + wn), int(0.02*wn+wn_locscale):(int(0.02*wn+wn_locscale) + wn), (int(center[2]-0.5*wn)):(int((center[2]-0.5*wn) + wn))];
    else:
        noiseMap = emmap[int(boxCoord[0]-0.5*wn + wn_locscale):(int(boxCoord[0]-0.5*wn + wn_locscale) + wn), int(boxCoord[1]-0.5*wn+ wn_locscale):(int(boxCoord[1]-0.5*wn + wn_locscale) + wn), (int(boxCoord[2]-0.5*wn + wn_locscale)):(int((boxCoord[2]-0.5*wn + wn_locscale)+wn))];

    #prepare noise map for scaling
    frequencyMap_noise = calculate_frequency_map(noiseMap);
    noiseMapFFT = np.fft.rfftn(noiseMap);
    noise_profile, frequencies_noise = compute_radial_profile(noiseMapFFT, frequencyMap_noise);

    #prepare windows of particle for scaling
    frequencyMap_mapWindow = calculate_frequency_map(np.zeros((wn_locscale, wn_locscale, wn_locscale)));

    for k, j, i in (masked_xyz_locs - wn_locscale / 2.0):
		
        emmap_wn = emmap[k: k+wn_locscale, j: j+wn_locscale, i: i+ wn_locscale];
        modmap_wn = modmap[k: k+wn_locscale, j: j+wn_locscale, i: i+ wn_locscale];
	    
        #do sharpening of the sliding window
        emmap_wn_FFT = np.fft.rfftn(np.copy(emmap_wn));
        modmap_wn_FFT = np.fft.rfftn(np.copy(modmap_wn));

        em_profile, frequencies_map = compute_radial_profile(emmap_wn_FFT, frequencyMap_mapWindow);
        mod_profile, _ = compute_radial_profile(modmap_wn_FFT, frequencyMap_mapWindow);
        scale_factors = compute_scale_factors(em_profile, mod_profile);
        map_b_sharpened, map_b_sharpened_FFT = set_radial_profile(emmap_wn_FFT, scale_factors, frequencies_map, frequencyMap_mapWindow);
	  
        #do interpolation of sharpening factors
        scale_factors_noise = np.interp(frequencies_noise, frequencies_map, scale_factors);

        #scale noise window with the interpolated scaling factors
        mapNoise_sharpened, mapNoise_sharpened_FFT = set_radial_profile(np.copy(noiseMapFFT), scale_factors_noise, frequencies_noise, frequencyMap_noise);

        #local filtering routines
        if locFilt == True:
            tmpRes = round(apix/locResMap[k, j, i], 3);
            				
            mapNoise_sharpened = lowPassFilter(mapNoise_sharpened_FFT, frequencyMap_noise, tmpRes);
            map_b_sharpened = lowPassFilter(map_b_sharpened_FFT, frequencyMap_mapWindow, tmpRes);
            
            #calculate noise statistics	  
            map_noise_sharpened_data = mapNoise_sharpened;	   
          	
            if ecdfBool:
                tmpECDF, sampleSort = estimateECDFFromMap(map_noise_sharpened_data, -1, -1);
                ecdf = np.interp(map_b_sharpened[central_pix, central_pix, central_pix], sampleSort, tmpECDF, left=0.0, right=1.0); 
            else:
                ecdf = 0;

            mean = np.mean(map_noise_sharpened_data);
            var = np.var(map_noise_sharpened_data);

            if var < 0.05:
                var = 0.05;
                mean = 0.0;
            if tmpRes == round(apix/100.0, 3):  
                mean = 0.0;
                var = 0.0;
                ecdf = 0;   	 			
        else:    
            #calculate noise statistics
            map_noise_sharpened_data = np.copy(mapNoise_sharpened);
         
            if ecdfBool:
                tmpECDF, sampleSort = estimateECDFFromMap(map_noise_sharpened_data, -1, -1);
                ecdf = np.interp(map_b_sharpened[central_pix, central_pix, central_pix], sampleSort, tmpECDF, left=0.0, right=1.0); 
            else:
                ecdf = 0;
            			
            mean = np.mean(map_noise_sharpened_data);
            var = np.var(map_noise_sharpened_data);
            if var < 0.05:
                var = 0.05;
                mean = 0.0;

        #append values to the sharpened values
        sharpened_vals.append(map_b_sharpened[central_pix, central_pix, central_pix]);
        sharpened_mean_vals.append(mean);
        sharpened_var_vals.append(var);     
        sharpened_ecdf_vals.append(ecdf);
       
    return np.array(sharpened_vals, dtype=np.float32), np.array(sharpened_mean_vals, dtype=np.float32), np.array(sharpened_var_vals, dtype=np.float32), np.array(sharpened_ecdf_vals, dtype=np.float32);

def put_scaled_voxels_back_in_original_volume_including_padding(sharpened_vals, masked_indices, map_shape):
    map_scaled = np.zeros(np.prod(map_shape))
    map_scaled[masked_indices] = sharpened_vals
    map_scaled = map_scaled.reshape(map_shape)
        
    return map_scaled;

def run_window_function_including_scaling(emmap, modmap, mask, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdfBool):
    """
    >>> emmap, modmap, mask = setup_test_data()
    >>> scaled_vol = run_window_function_including_scaling(emmap,modmap,mask,wn=10,apix=1.0)
    >>> np.copy(EMNumPy.em2numpy(scaled_vol))[scaled_vol.get_xsize() / 2][scaled_vol.get_ysize() / 2]
    array([ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
            0.12524424,  0.15562208,  0.18547297,  0.24380369,  0.31203741,
            0.46546721,  0.47914436,  0.31334871,  0.28510684,  0.21345402,
            0.17892323,  0.        ,  0.        ,  0.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.        ,  0.        ], dtype=float32)
    """
    masked_xyz_locs, masked_indices, map_shape = get_xyz_locs_and_indices_after_edge_cropping_and_masking(mask, wn_locscale);

    sharpened_vals, sharpened_mean_vals, sharpened_var_vals, sharpened_ecdf_vals = get_central_scaled_pixel_vals_after_scaling(emmap, modmap, masked_xyz_locs, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdfBool);
     
    map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(sharpened_vals, masked_indices, map_shape);
    mean_map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(sharpened_mean_vals, masked_indices, map_shape);	 
    var_map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(sharpened_var_vals, masked_indices, map_shape);
    ecdf_map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(sharpened_ecdf_vals, masked_indices, map_shape);

    return map_scaled, mean_map_scaled, var_map_scaled, ecdf_map_scaled; 
   
def split_sequence_evenly(seq, size):
    """
    >>> split_sequence_evenly(list(range(9)), 4)
    [[0, 1], [2, 3, 4], [5, 6], [7, 8]]
    >>> split_sequence_evenly(list(range(18)), 4)
    [[0, 1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12, 13], [14, 15, 16, 17]]
    """
    newseq = []
    splitsize = 1.0 / size * len(seq)
    for i in range(size):
        newseq.append(seq[int(round(i * splitsize)):int(round((i + 1) * splitsize))])
    return newseq
    
def merge_sequence_of_sequences(seq):
    """
    >>> merge_sequence_of_sequences([list(range(9)), list(range(3))])
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 0, 1, 2]
    >>> merge_sequence_of_sequences([list(range(9)), [], list(range(3))])
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 0, 1, 2]
    """
    newseq = [number for sequence in seq for number in sequence]
    
    return newseq
   

def run_window_function_including_scaling_mpi(emmap, modmap, mask, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdfBool):
    """
    >>> emmap_name, modmap_name, mask_name = setup_test_data_to_files()
    >>> import subprocess
    >>> n = subprocess.call(mpi_cmd.split())
    >>> scaled_vol = get_image('scaled.mrc')
    >>> np.copy(EMNumPy.em2numpy(scaled_vol))[scaled_vol.get_xsize() / 2][scaled_vol.get_ysize() / 2]
    array([ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
            0.12524424,  0.15562208,  0.18547297,  0.24380369,  0.31203741,
            0.46546721,  0.47914436,  0.31334871,  0.28510684,  0.21345402,
            0.17892323,  0.        ,  0.        ,  0.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.        ,  0.        ], dtype=float32)
    >>> n = [os.remove(each_file) for each_file in [emmap_name, modmap_name, mask_name, 'scaled.mrc']]
    """
    from mpi4py import MPI;
    comm = MPI.COMM_WORLD;
    rank = comm.Get_rank();
    size = comm.Get_size();
         
    if rank == 0:

        print("*****************************")
        print("********* LocScale **********")
        print("*****************************")

        masked_xyz_locs, masked_indices, map_shape = \
        get_xyz_locs_and_indices_after_edge_cropping_and_masking(mask, wn);
         
        zs, ys, xs = masked_xyz_locs.T;
        zs = split_sequence_evenly(zs, size);
        ys = split_sequence_evenly(ys, size);
        xs = split_sequence_evenly(xs, size);
    else:
        zs = None;
        ys = None;
        xs = None;
     
    zs = comm.scatter(zs, root=0);
    ys = comm.scatter(ys, root=0);
    xs = comm.scatter(xs, root=0);
 
    masked_xyz_locs = np.column_stack((zs, ys, xs));
 
    sharpened_vals, sharpened_mean_vals, sharpened_var_vals, sharpened_ecdf_vals = get_central_scaled_pixel_vals_after_scaling(emmap, modmap, masked_xyz_locs, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdfBool);

    comm.barrier();
    sharpened_vals = comm.gather(sharpened_vals, root=0);
    sharpened_mean_vals = comm.gather(sharpened_mean_vals, root=0);
    sharpened_var_vals = comm.gather(sharpened_var_vals, root=0);		
    sharpened_ecdf_vals = comm.gather(sharpened_ecdf_vals, root=0);
     
    if rank == 0:
        sharpened_vals = merge_sequence_of_sequences(sharpened_vals);
        sharpened_mean_vals = merge_sequence_of_sequences(sharpened_mean_vals);
        sharpened_var_vals = merge_sequence_of_sequences(sharpened_var_vals);
        sharpened_ecdf_vals = merge_sequence_of_sequences(sharpened_ecdf_vals);

        map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(np.array(sharpened_vals),
        masked_indices, map_shape);

        mean_map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(np.array(sharpened_mean_vals),
        masked_indices, map_shape);

        var_map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(np.array(sharpened_var_vals),
        masked_indices, map_shape);
        
        ecdf_map_scaled = put_scaled_voxels_back_in_original_volume_including_padding(np.array(sharpened_ecdf_vals),
        masked_indices, map_shape);

    else:
        map_scaled = None;
        mean_map_scaled = None;
        var_map_scaled = None;		
        ecdf_map_scaled = None;

    comm.barrier();		

    return map_scaled, mean_map_scaled, var_map_scaled, ecdf_map_scaled, rank;
  
def write_out_final_volume_window_back_if_required(wn, window_bleed_and_pad, Volume):
    
    if window_bleed_and_pad:
        map_shape = [(Volume.shape[0] - wn), (Volume.shape[1] - wn), (Volume.shape[2] - wn)]
        Volume = pad_or_crop_volume(Volume, (map_shape))
                                         
    return Volume;

def launch_amplitude_scaling(em_map, model_map, apix, stepSize, wn_locscale, wn, method, locResMap, noiseBox, mpi, ecdf ):

    startTime = time.time();
    emmap, modmap, mask, wn, wn_locscale, window_bleed_and_pad, method, locFilt, locResMap, boxCoord = prepare_mask_and_maps_for_scaling(em_map, model_map, apix, wn_locscale, wn, method, locResMap, noiseBox); 
    meanNoise, varNoise, sample = FDRutil.estimateNoiseFromMap(emmap, wn, boxCoord);

    if not mpi:
        stepSize = int(stepSize);
        if stepSize == 1:
            LocScaleVol, meanVol, varVol, ecdfVol = run_window_function_including_scaling(emmap, modmap, mask, wn, wn_locscale , apix, locFilt, locResMap, boxCoord, ecdf);
        elif stepSize <= 0:
            print("Invalid step size parameter. It has to be greater than 0! Quit program ...");
            return;
        else:
            LocScaleVol, meanVol, varVol, ecdfVol = calculate_scaled_map(emmap, modmap, mask, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdf, stepSize);
        
        print("Local amplitude scaling finished ...")

        LocScaleVol = mask*LocScaleVol;

        if not ecdf:
            ecdfVol = 0;
        else:
            ecdfVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, ecdfVol);

        LocScaleVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, LocScaleVol);
        meanVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, meanVol);	
        varVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, varVol);
        
        return LocScaleVol, meanVol, varVol, ecdfVol;

        #qVol = calcQMap(LocScaleVol, meanVol, varVol, ecdfVol, 0, 0, mask, method, testProc);
        #qVol = np.subtract(np.ones(qVol.shape), qVol);

        #write the volumes		
        #LocScaleVol = write_out_final_volume_window_back_if_required(args, wn_locscale, window_bleed_and_pad, LocScaleVol, splitFilename[0] + '_scaled.mrc');
        #qVol = write_out_final_volume_window_back_if_required(args, wn_locscale, window_bleed_and_pad, qVol, splitFilename[0] + '_confidenceMap.mrc');

        #endTime = time.time()
        #runTime = endTime - startTime
    
    elif mpi:
        LocScaleVol, meanVol, varVol, ecdfVol, rank = run_window_function_including_scaling_mpi(emmap, modmap, mask, wn, wn_locscale, apix, locFilt, locResMap, boxCoord, ecdf);       
        if rank == 0:
            print("Local amplitude scaling finished ...")			

            if not ecdf:
                ecdfVol = 0;
            else:
                ecdfVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, ecdfVol);

            LocScaleVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, LocScaleVol);
            meanVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, meanVol);
            varVol = write_out_final_volume_window_back_if_required(wn_locscale, window_bleed_and_pad, varVol);

            return LocScaleVol, meanVol, varVol, ecdfVol;
	


            #qVol = calcQMap(LocScaleVol, meanVol, varVol, ecdfVol, 0, 0, mask, method, testProc);
            #qVol = np.subtract(np.ones(qVol.shape), qVol);

            #write the volumes
            #LocScaleVol = write_out_final_volume_window_back_if_required(args, wn_locscale, window_bleed_and_pad, LocScaleVol, splitFilename[0] + '_scaled.mrc')
            #qVol = write_out_final_volume_window_back_if_required(args, wn_locscale, window_bleed_and_pad, qVol, splitFilename[0] + '_confidenceMap.mrc')

            #endTime = time.time()
            #runTime = endTime - startTime;
            #pp = makeDiagnosticPlot(emmap, wn, wn_locscale, True, boxCoord);
            #pp.savefig("diag_image.pdf");
            #pp.close();
            #printSummary(args, runTime);




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 20 14:56:33 2022

@author: alok
"""
import mrcfile
import numpy as np

def compute_radial_profile_proper(vol, frequency_map):

    vol_fft = np.fft.rfftn(vol, norm="ortho");
    dim = vol_fft.shape;
    ps = np.real(np.abs(vol_fft));
    frequencies = np.fft.rfftfreq(dim[0])[1:];
    #bins = np.digitize(frequency_map, frequencies);
    #bins = bins - 1;
    x, y, z = np.indices(ps.shape)
    radii = np.sqrt(x**2 + y**2 + z**2)
    radii = radii.astype(int)
    radial_profile = np.bincount(radii.ravel(), ps.ravel()) / np.bincount(radii.ravel())
    radial_profile = radial_profile[1:int(ps.shape[0]/2)+1]

    return radial_profile, frequencies;

def compute_radial_profile_simple(vol, return_frequencies=False):
    from emmer.include.confidenceMapUtil import FDRutil
    frequency_map = FDRutil.calculate_frequency_map(np.zeros(vol.shape))
    
    em_profile, frequencies_map = compute_radial_profile_proper(vol, frequency_map)
    
    if return_frequencies:
        return em_profile, frequencies_map
    else:
        return em_profile
    
def set_radial_profile_proper(vol, scale_factors, frequencies, frequency_map, shape):
    vol_fft = np.fft.rfftn(np.copy(vol), norm='ortho');
    scaling_map = np.interp(frequency_map, frequencies, scale_factors);
    scaled_map_fft = scaling_map * vol_fft;
    scaled_map = np.real(np.fft.irfftn(scaled_map_fft, shape, norm='ortho'));

    return scaled_map, scaled_map_fft;


def set_radial_profile_simple(vol, scale_factors, frequencies):
    from emmer.include.confidenceMapUtil import FDRutil
    frequency_map = FDRutil.calculate_frequency_map(np.zeros(vol.shape))
    
    map_shape = vol.shape
    map_b_sharpened, _ = set_radial_profile_proper(vol, scale_factors, frequencies, frequency_map, map_shape);
    
    return map_b_sharpened



def compute_scale_factors(em_profile, ref_profile):
    scale_factor = np.sqrt(ref_profile**2/em_profile**2)
    return scale_factor






## Script to add two half maps and save the result as a mrc file if the output filename is specified

def average_half_maps(halfmap_1_path, halfmap_2_path, output_filename, return_map=False, fsc_filter=True):
    '''
    Function to add two half maps

    Parameters
    ----------
    halfmap_1_path : str 
        Path to halfmap 1
        
    halfmap_2_path : str 
        Path to halfmap 2
        

    Returns
    -------
    output_filename : str
        Path to output file
    


    '''
    from emmer.ndimage.map_utils import save_as_mrc, load_map
    from emmer.ndimage.filter.apply_fsc_filter import apply_fsc_filter
    halfmap1, apix = load_map(halfmap_1_path)
    halfmap2, _ = load_map(halfmap_2_path)
    
    assert halfmap1.shape == halfmap2.shape
    
    full_map = (halfmap1 + halfmap2) / 2

    # Apply FSC filter if specified
    if fsc_filter:
        full_map = apply_fsc_filter(full_map, apix, halfmap_1=halfmap1, halfmap_2=halfmap2)

    if output_filename is not None:
        save_as_mrc(full_map, output_filename, apix, verbose=True)
    
    if return_map:
        return full_map
    else:   
        return output_filename
    



## Script to estimate global bfactor of a cryo EM map using the radial profile

def estimate_bfactor_map(emmap_path, wilson_cutoff, fsc_cutoff, return_fit=False):
    '''
    Script to estimate the global bfactor of a cryo EM density map using the radial profile.
    emmap_path: path to the cryo EM density map
    wilson_cutoff: the Wilson score cutoff for the global bfactor estimation
    fsc_cutoff: the FSC cutoff for the global bfactor estimation
    return_fit: whether to return the fit quality of the bfactor (True)
    returns:
    bfactor: the estimated bfactor (float)
    '''
    from emmer.ndimage.map_utils import load_map
    from emmer.ndimage.radial_profile.frequency_array import frequency_array 
    from emmer.ndimage.radial_profile.compute_radial_profile import compute_radial_profile
    from emmer.ndimage.radial_profile.estimate_bfactor_pwlf import estimate_bfactor_pwlf
    
    emmap, apix = load_map(emmap_path)

    rp_unsharp = compute_radial_profile(emmap)
    freq = frequency_array(amplitudes=rp_unsharp, apix=apix)
        
    bfactor,_,(fit,z,slopes) = estimate_bfactor_pwlf(freq,rp_unsharp, 
                                                             wilson_cutoff=wilson_cutoff, 
                                                             fsc_cutoff=fsc_cutoff)
    
    if return_fit:
        return bfactor, fit
    else:
        return bfactor


# -*- coding: utf-8 -*-
"""
Created on Sun Jul 18 14:52:11 2021
"""

# get_dssp contains several functions to analyse secondary structure
# information in PDB structures using Biopython

# global imports
from Bio.PDB import PDBParser
from Bio.PDB.DSSP import DSSP
from Bio.PDB import PDBList
from emmer.pdb.pdb_utils import detect_pdb_input
from emmer.pdb.SSE.get_dssp import get_dssp
#%% functions

def get_secondary_structure_residues_content(pdbid,pdb_path):
     '''
     Input: pdbid: string '3j5p' 
     pdb_path: string - gives the full path of the PDB file
     '''
     # local imports
     import os

     
     dssp = get_dssp(pdbid,pdb_path)    
     secondary_structures = [dssp[key][2] for key in dssp.keys()]
     helix_residues = secondary_structures.count('H') + secondary_structures.count('G') + secondary_structures.count('I')
     sheet_residues = secondary_structures.count('B') + secondary_structures.count('E')
     loop_residues = secondary_structures.count('T') + secondary_structures.count('S')
     total_residues = len(secondary_structures)
     
     secondary_structure_distribution = {}
     secondary_structure_distribution['helix'] = helix_residues/total_residues
     secondary_structure_distribution['sheet'] = sheet_residues/total_residues
     secondary_structure_distribution['loop'] = loop_residues/total_residues
     secondary_structure_distribution['total'] = total_residues
     

     return secondary_structures,secondary_structure_distribution
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 20 17:10:49 2022

@author: alok
"""
import numpy as np

def mesh_surface_area(data, threshold, apix):
    from skimage import measure
    
    verts, faces,_,_ = measure.marching_cubes(data, threshold)
    surface_area = measure.mesh_surface_area(verts, faces) * apix**2
    return surface_area









import setuptools
from setuptools import setup
import pathlib
emmer_path = pathlib.Path(__file__).parent.resolve()
long_description = (emmer_path / "README.md").read_text()


setup(name='emmer',
    version='0.1.2',
    author='Alok Bharadwaj, Maarten Joosten, Stefan T Huber, Arjen Jakobi, Reinier de Bruin',
    url='https://gitlab.tudelft.nl/aj-lab/emmer',
    description= "A python toolkit for the cryo-EM developer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='3-clause BSD',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'emmer': ['include/geckodriver','include/*.pickle','*.pickle', '*.pdb','*.mrc','*.cif','*.npy']},
    install_requires=['numpy>=1.19.5','scipy>=1.5.4','pandas>=1.1.5','mrcfile>=1.3.0','gemmi>=0.4.8',\
                    'pypdb>=2.0','scikit-learn>=0.0','pwlf>=2.0.4','tqdm>=4.62.3','more_itertools>=8.10.0',\
                    'scikit-image>=0.17.2','biopython==1.78','matplotlib==3.5.1', 'coverage','pyfiglet>=0.8.post1', 'wget>=3.2'],
    entry_points={
        'console_scripts': [
            'emmer=emmer.main:main',
        ],
    },
    zip_safe= False)




## Script to extract map intensities around atoms between a given residue range

def trim_map_between_residues(emmap_path, pdb_path, chain_name, residue_range=None, dilation_radius=3, verbose=False):
    '''
    Function to extract map intensities around atoms between a given residue range

    Parameters
    ----------
    emmap_path : str
        Path to a map file 
    pdb_path : str
        Path to a PDB/MMCIF file
    chain_name : str
        Chain name
    residue_range : list, optional
        To extract all atoms between residue id
        residue_range=[start_res_id, end_res_id] (both incl). The default is [0,-1], which returns all residues present in a chain. 
    dilation_radius : float, optional
        The radius of the sphere (in Ang) to place at atomic positions determined by the PDB file. Default is 3A.

    Returns
    -------
    trimmed_map : numpy.ndarray
    

    '''
    import numpy as np
    from emmer.pdb.pdb_utils import detect_pdb_input
    from emmer.pdb.pdb_utils import get_atomic_positions_between_residues
    from emmer.pdb.convert.convert_pdb_to_mrc_position import convert_pdb_to_mrc_position
    from emmer.ndimage.map_utils import dilate_mask, load_map
    
    emmap, apix = load_map(emmap_path)
        
    map_shape = emmap.shape
    
    mask = np.zeros(map_shape)
    
    gemmi_st = detect_pdb_input(pdb_path)
    
    pdb_positions = get_atomic_positions_between_residues(gemmi_st, chain_name, residue_range)
    
    if verbose:
        print("Found {} atom sites".format(len(pdb_positions)))
    
    mrc_position = convert_pdb_to_mrc_position(pdb_positions, apix)
    zz,yy,xx = zip(*mrc_position)
    mask[zz,yy,xx] = 1
    
    #dilation_radius = 3 #A
    dilation_radius_int = round(dilation_radius / apix)
    dilated_mask = dilate_mask(mask, radius=dilation_radius_int)
    
    trimmed_map = emmap * dilated_mask
    
    return trimmed_map



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 22 15:23:14 2021

@author: alok
"""

import numpy as np
  
def get_index_at_resolution(freq, resolution):
    probe_freq = 1/resolution
    index_at_resolution = np.argmin(abs(freq-probe_freq))
    return index_at_resolution

def crop_profile_between_resolution(freq, amplitude, low_resolution, high_resolution):
    ## Check if inverse of minimum resolution and inverse of maximum resolution are within freq range
    if 1/low_resolution < freq[0] or 1/high_resolution > freq[-1]:
        raise UserWarning("Minimum resolution or maximum resolution is out of range")
    
    low_frequency_index = get_index_at_resolution(freq, low_resolution)
    high_frequency_index = get_index_at_resolution(freq, high_resolution)

        
    crop_freq = freq[low_frequency_index:high_frequency_index]
    crop_amplitude = amplitude[low_frequency_index:high_frequency_index]
    
    return crop_freq, crop_amplitude
       
def resample_1d(x_old,y_old,num,xlims=None):
    '''
    Sample an given x-y data in a new grid 

    Parameters
    ----------
    x_old : numpy.ndarray
        data in x axis (same dim as y_old)
    y_old : numpy.ndarray
        data in y axis (same dim as x_old)
    num : int
        new number of data points

    Returns
    -------
    x_new : numpy.ndarray
        resampled x axis
    y_new : numpy.ndarray
        resampled y axis

    '''
    from scipy.interpolate import interp1d
    
    f = interp1d(x_old, y_old,kind='slinear',fill_value='extrapolate')
    if xlims is None:
        x_new = np.linspace(x_old[0], x_old[-1], num=num)
    else:
        xmin = xlims[0]
        xmax = xlims[1]
        x_new = np.linspace(xmin, xmax,num=num)
        
    y_new = f(x_new)
    #y_new[y_new>1]=1
    return x_new, y_new
  


def measure_wilson_deviation(freq, amplitudes, wilson_cutoff, fsc_cutoff, wilson_range=None, return_exponential_fit=False):
    '''
    Function to measure the deviations in wilson region from an exponential behaviour
    '''
    import numpy as np
    ## import estimate_bfactor_standard
    from emmer.ndimage.radial_profile.estimate_bfactor_standard import estimate_bfactor_standard
    if wilson_range is None:
        wilson_range_start = wilson_cutoff
        wilson_range_end = fsc_cutoff
    else:
        low_freq_cutoff_wilson = wilson_range.max()
        high_freq_cutoff_wilson = wilson_range.min()
        assert low_freq_cutoff_wilson > wilson_cutoff and low_freq_cutoff_wilson < 1/freq[0]
        assert high_freq_cutoff_wilson < fsc_cutoff and high_freq_cutoff_wilson < 1/freq[-1]
        
        wilson_range_start = low_freq_cutoff_wilson
        wilson_range_end = high_freq_cutoff_wilson
        
    bfactor, amp = estimate_bfactor_standard(freq, amplitudes, wilson_cutoff=wilson_cutoff, fsc_cutoff=fsc_cutoff, 
                                             return_amplitude=True)
    
    exponential_fit = amp * np.exp(bfactor * 0.25 * freq**2)
    
    deviations = amplitudes - exponential_fit
    
    deviation_start_index = get_index_at_resolution(freq,wilson_range_start)
    deviation_end_index = get_index_at_resolution(freq,wilson_range_end)
    
    deviations[:deviation_start_index] = 0
    deviations[deviation_end_index:] = 0
    
    if return_exponential_fit:
        return deviations, exponential_fit
    else:
        return deviations
    
def get_average_profile(profile_list, return_variance=False):
    profile_list = np.array(profile_list)
    average_profile = np.einsum("ij->j", profile_list) / len(profile_list)
    
    variation = []
    for col_index in range(profile_list.shape[1]):
        col_extract = profile_list[:,col_index]
        variation.append(col_extract.std())

    variation = np.array(variation)
    
    if return_variance:
        return average_profile, variation
    else:
        return average_profile




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec  3 14:11:29 2020

@author: Alok, Arjen, Maarten, Stefan, Reinier 
"""

# pdb_tools contains several useful fucntions for common manipulations with
# pdb structures, making use of the gemmi package. functions are classified 
# into pdb_tools when they can be considered an application on their own
# but do not have so many distinct features that they warrent their own script.

# global imports
import mrcfile
import gemmi
import numpy as np

#%% functions

def compute_rmsd_two_pdb(input_pdb_1, input_pdb_2, use_gemmi_structure=True, return_array=False):
    from emmer.pdb.pdb_utils import detect_pdb_input, get_all_atomic_positions
    from scipy.spatial import distance
    
    
    if use_gemmi_structure:
        st_1 = detect_pdb_input(input_pdb_1)
        st_2 = detect_pdb_input(input_pdb_2)
        
        num_atoms_1 = st_1[0].count_atom_sites()
        num_atoms_2 = st_2[0].count_atom_sites()
        
        positions_1 = get_all_atomic_positions(st_1)
        positions_2 = get_all_atomic_positions(st_2)
        
        assert num_atoms_1 == num_atoms_2
    
    else:
        positions_1 = input_pdb_1
        positions_2 = input_pdb_2
        
        assert len(positions_1) == len(positions_2)
        
    
    atomic_distance = []
    for index in range(len(positions_1)):
        dist = distance.euclidean(positions_1[index], positions_2[index])
        atomic_distance.append(dist)
    
    atomic_distance = np.array(atomic_distance)
    
    if return_array:
        return atomic_distance
    else:
        return np.mean(atomic_distance)



# -*- coding: utf-8 -*-
"""
Created on Sun Jul 18 14:08:55 2021
"""

# pdb_util contains helper functions used in other applications in the emmer
# toolbox. pdb_util functions are classified as such if there is little need
# to manually call these functions and/or when their output is used in
# many different applications.

# global imports
import gemmi
import numpy as np
#%% functions

def get_gemmi_st_from_id(pdb_id):
    '''
    Returns a gemmi.Structure() containing the PDB coordinate information and headers for a given PDB-ID. 

    Parameters
    ----------
    pdb_id : string
        PDB ID: "3j5p"

    Returns
    -------
    gemmi_st : gemmi.Structure()

    '''          
    import pypdb
    
    try: 
        pdb_file = pypdb.get_pdb_file(pdb_id,filetype='pdb',compression=False)
        pdb_struc = gemmi.read_pdb_string(pdb_file)
        print("- Successfully downloaded PDBgemmi {} from database".format(pdb_id))
        return pdb_struc
    except AttributeError:
        cif_file = pypdb.get_pdb_file(pdb_id, filetype='cif',\
                                          compression=False)
        cif_struc = gemmi.read_pdb_string(cif_file)
        return cif_struc
    
def check_if_gemmi_st_is_downloadable(pdb_id):
    """[summary]

    Args:
        pdb_id (str): PDB ID, like: "3j5p"

    Returns:
        Bool: indicating whether the gemmi structure is downloadable (True) or not (False)
    """    
    import pypdb
    
    try: 
        pdb_file = pypdb.get_pdb_file(pdb_id, filetype='pdb',compression=False)
        return True
    except AttributeError:
        try:
            cif_file = pypdb.get_pdb_file(pdb_id, filetype='cif', compression=False)
            return True
        except Exception as e:
            print("- Exception: {}".format(e))
            return False
        
def detect_pdb_input(input_pdb):
    '''
    Function to detect the type of input the user has passed and return gemmi.Structure as output

    Parameters
    ----------
    input_pdb : str or gemmi.Structure
        Input can be either 
        a) string: either (i) pdb path, for ex: "path/to/pdb.pdb" or 
                          (ii)pdb_id "3j5p"
        b) gemmi.Structure()

    Returns
    -------
    pdb_structure : gemmi.Structure()
        Parsed input of type gemmi.Structure()
        Returns nothing if input cannot be parsed properly.

    '''
    from emmer.pdb.pdb_utils import get_gemmi_st_from_id
    
    if isinstance(input_pdb, str):
        if input_pdb.split(sep='.')[-1] in ['pdb','cif']:  # Then input is a file path
            pdb_structure = gemmi.read_structure(input_pdb)
            return pdb_structure
        elif ("/" not in input_pdb) or ("\\" not in input_pdb) or ("." not in input_pdb): # Then input is not file path but pdb_id
            pdb_structure = get_gemmi_st_from_id(input_pdb)
            return pdb_structure
        else:
            print("Input cannot be parsed. Please pass a pdb_id (as string) or pdb_path (as string)")
    elif isinstance(input_pdb, gemmi.Structure):
        pdb_structure = input_pdb.clone()
        return pdb_structure
    else:
        print("Unknown datatype for input. Please pass either a gemmi.Structure() or a\
              string (pointing to pdb_path or pdb_id")
    
def set_to_center_of_unit_cell(input_pdb, unitcell):
    '''
    Function to set the center of mass of a PDB structure to the center of a unitcell

    Parameters
    ----------
    pdb_structure : gemmi.Structure
        Input structure 
    unitcell : gemmi.UnitCell
        Input unitcell

    Returns
    -------
    centered_pdb : gemmi.Structure

    '''
    from emmer.pdb.pdb_utils import shift_coordinates
    
    pdb_structure_local = detect_pdb_input(input_pdb)
    center_of_mass_old = np.array(pdb_structure_local[0].calculate_center_of_mass().tolist())
    center_of_mass_new = np.array([unitcell.a/2, unitcell.b/2, unitcell.c/2])
    
    translation_matrix = center_of_mass_new - center_of_mass_old
    shifted_structure = shift_coordinates(trans_matrix=translation_matrix, input_pdb=pdb_structure_local)
    
    return shifted_structure

def get_atomic_positions_between_residues(input_pdb, chain_name, res_range = None, model_index=0):
    '''
    Extract atom positions between residue range

    Parameters
    ----------
    gemmi_structure : gemmi.Structure()
        input gemmi structure
    chain_name : str
        
    res_range : list
        res_range = [start_res, end_res] (both incl)

    Returns
    -------
    pdb_positions : list
    
    pdb_positions = [[x1, y1, z1], [x2, y2, z3]...] (values in Angstorm)
    '''
    structure = detect_pdb_input(input_pdb)
    gemmi_model = structure[model_index]
    pdb_positions = []
    for chain in gemmi_model:
        if chain.name == chain_name:
            if res_range is not None:
                for res in chain:
                    if res.seqid.num >= res_range[0] and res.seqid.num <= res_range[1] :
                        for atom in res:
                            pdb_positions.append(atom.pos.tolist())
            else:
                for res in chain:
                    for atom in res:
                        pdb_positions.append(atom.pos.tolist())
                        
    
    return pdb_positions

def get_unit_cell_estimate(pdb_struct,vsize):
          
    '''
    Find an estimated size of unit cell in A based on nunber of atoms and apix

    As reference: PDB3J5P had ~18000 atoms and a box size of 256^3
          
    '''

    number_of_atoms = pdb_struct[0].count_atom_sites()
    estimated_box_size = number_of_atoms * 256 / 18000
    unit_cell_dim =  estimated_box_size * vsize
    unitcell = gemmi.UnitCell(unit_cell_dim,unit_cell_dim,unit_cell_dim,90,90,90)
          
    return unitcell

def replace_pdb_column_with_arrays(input_pdb, replace_column, replace_array):
    '''
    Replace a column in the PDB (either bfactor or occupancy) with values from an array where the \ 
        array index position correspods to atom location in the cra generator

    Parameters
    ----------
    input_pdb : str to gemmi structure
        path to pdb file or a gemmi structure.
    replace_column : str
        string to specify which PDB column to replace values with
    replace_array : numpy.ndarray (1d)
        replace values of column with these values

    Returns
    -------
    st: gemmi.Structure
    

    '''
    
    st = detect_pdb_input(input_pdb)
    
    replace_array = np.clip(replace_array, 0, 999)  ## PDB array only allows a maximum of three digits
    for i,cra_gen in enumerate(st[0].all()):
        if replace_column=="bfactor":
            cra_gen.atom.b_iso = replace_array[i]
        elif replace_column=="occ":
            cra_gen.atom.occ = replace_array[i]
        else:
            raise UserWarning("Please input either bfactor or occ for the replace_column variable")
        
    
    return st

def remove_charges(input_pdb):
    structure = detect_pdb_input(input_pdb)
    
    for model in structure:
        for chain in model:
            for residue in chain:
                for atom in residue:
                    atom.charge = 0
    
    return structure

def shift_coordinates(input_pdb, trans_matrix=[0,0,0]):
    """
    Shift atomic coordinates based on a translation matrix
    """
    structure = detect_pdb_input(input_pdb)
            
    for model in structure:
        for chain in model:
            for residue in chain:
                for atom in residue:
                    atom.pos = gemmi.Position(atom.pos.x+trans_matrix[0],
                                              atom.pos.y+trans_matrix[1],
                                              atom.pos.z+trans_matrix[2])
        
    return structure
  
def split_model_based_on_nucleotides(input_pdb):
    structure = detect_pdb_input(input_pdb)
    dna_model = gemmi.Model('D')
    rna_model = gemmi.Model('R')
    
    for model in structure:
        for chain in model:
            dna_model.add_chain(chain.name)
            rna_model.add_chain(chain.name)
            for res in chain:
                if res.name in ['C','G','A','U','I']:
                    rna_model[chain.name].add_residue(res)
                elif res.name in ['DC','DG','DA','DU','DI','DT']:
                    dna_model[chain.name].add_residue(res)
    
    dna_st = gemmi.Structure()
    dna_st.add_model(dna_model)
    
    rna_st = gemmi.Structure()
    rna_st.add_model(rna_model)

    return dna_st, rna_st    

def get_all_atomic_positions(input_pdb, as_dictionary=False):
    '''
    Extract atom positions

    Parameters
    ----------
    gemmi_structure : gemmi.Structure()
        input gemmi structure
    chain_name : str
        
    res_range : list
        res_range = [start_res, end_res] (both incl)

    Returns
    -------
    pdb_positions : list
    
    pdb_positions = [[x1, y1, z1], [x2, y2, z3]...] (values in Angstorm)
    '''
    
    st = detect_pdb_input(input_pdb)
    
    if as_dictionary:
        pdb_positions = {}
        for i,cra_obj in enumerate(st[0].all()):
            pdb_positions[i] = np.array(cra_obj.atom.pos.tolist())
        
        return pdb_positions
                        
    
    else:
        pdb_positions = []
        for chain in st[0]:
            for res in chain:
                for atom in res:
                    pdb_positions.append(atom.pos.tolist())
                            
        
        return np.array(pdb_positions)

def set_all_atomic_positions(input_pdb, positions_dictionary):
    '''
    Input a dictionary where keys are atomic "access IDs " generated by the function get_all_atomic_positions

    Parameters
    ----------
    gemmi_structure : TYPE
        DESCRIPTION.
    positions_dictionary : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    '''
    
    st = detect_pdb_input(input_pdb)
    for i, cra_obj in enumerate(st[0].all()):
        new_position = gemmi.Position(positions_dictionary[i][0],positions_dictionary[i][1],positions_dictionary[i][2])
        cra_obj.atom.pos = new_position
    
    return st
    

def get_bfactors(input_pdb, model_index=None, return_type="list"):
    """
    Get B-factors of atoms
    """
    dict_chain = {}
    if model_index is not None:
        input_structure = detect_pdb_input(input_pdb)
        probe_model = input_structure[model_index]
        structure = gemmi.Structure()
        structure.add_model(probe_model)
    else:
        structure = detect_pdb_input(input_pdb)
    
    
        
    list_bfact = []
    for model in structure:
        for chain in model:
            if not chain.name in dict_chain:
                dict_chain[chain.name] = {}
            for residue in chain:
                residue_id = str(residue.seqid.num)+'_'+residue.name
                for atom in residue:
                    list_bfact.append(atom.b_iso)
                avg_bfact = sum(list_bfact)/float(len(list_bfact))
                dict_chain[chain.name][residue_id] = round(avg_bfact,3)
    
    if return_type=="list":
        return list_bfact
    elif return_type=="dictionary":
        return dict_chain
    else:
        raise UserWarning("Enter return_type as either list or dictionary")
        

def add_atomic_bfactors(input_pdb, additional_biso):
    '''
    Function to modify atomic bfactors uniformly by adding or subtracting b factors to each atom present in the PDB.
    

    Parameters
    ----------
    input_pdb : string or gemmi.Structure
        Path to a PDB file or gemmi structure
    
    additional_biso : float, 
        Parameter to specify how the bfactors of the atomic model be modified

    Returns
    -------
     st : gemmi.Structure 

    '''
    
    structure = detect_pdb_input(input_pdb)
    
    for model in structure:
        for chain in model:
            for res in chain:
                for atom in res:
                    atom.b_iso += additional_biso
    
    
    return structure
    
def set_atomic_bfactors(input_pdb, b_iso=None):
    '''
    Function to modify atomic bfactors uniformly by adding or subtracting b factors to each atom present in the PDB.
    

    Parameters
    ----------
    in_model_path : str, optional
        Path to a PDB file. 
    gemmi_st : gemmi.Structure()
        Pass a gemmi.Structure() instead of a path to perform computation online
    b_iso : float, 
        Parameter to specify the bfactors of the atomic model

    Returns
    -------
    If in_model_path is passed, returns the output model path.. Else returns the gemmi.Structure()

    '''
    structure = detect_pdb_input(input_pdb)

    for model in structure:
        for chain in model:
            for res in chain:
                for atom in res:
                    atom.b_iso = b_iso

    return structure

def get_residue_ca_coordinates(input_pdb):
    
    structure = detect_pdb_input(input_pdb)
    
    dict_coord = {}
    for model in structure:
        if not model.name in dict_coord: dict_coord[model.name] = {}
        for chain in model:
            polymer = chain.get_polymer()
            #skip non polymers
            #if not polymer: continue
            
            if not chain.name in dict_coord[model.name]: 
                dict_coord[model.name][chain.name] = {}
            for residue in chain:
                residue_id = str(residue.seqid.num)+'_'+residue.name
                residue_centre = ()
                if residue.name in ['A','T','C','G','U']:#nuc acid
                    for atom in residue:
                        if atom.name in ["P","C3'","C1'"]:
                            residue_centre = (atom.pos.x,atom.pos.y,atom.pos.z)
                else:
                    for atom in residue:
                        if atom.name == 'CA':#prot
                            residue_centre = (atom.pos.x,atom.pos.y,atom.pos.z)
                if len(residue_centre) == 0:#non nuc acid / prot
                    try: 
                        center_index = len(residue)/2
                        atom = residue[center_index]
                        residue_centre = (atom.pos.x,atom.pos.y,atom.pos.z)
                    except: 
                        for atom in residue:
                            residue_centre = (atom.pos.x,atom.pos.y,atom.pos.z)
                            break #first atom
                if len(residue_centre) > 0:
                    dict_coord[model.name][str(chain.name)][str(residue.seqid.num)] = \
                                            [residue_centre, residue.name]

    return dict_coord

def get_coordinates(input_pdb):
    list_coord = []
    structure = detect_pdb_input(input_pdb)
    for model in structure:
        for chain in model:
            for residue in chain:
                for atom in residue:
                    coord = atom.pos #gemmi Position
                    list_coord.append([coord.x,coord.y,coord.z])
    return list_coord

def find_number_of_neighbors(input_pdb, atomic_position, window_size_A):    
    gemmi_structure = detect_pdb_input(input_pdb)
    
    # Neighbor Search initialize
    
    ns = gemmi.NeighborSearch(gemmi_structure[0], gemmi_structure.cell, window_size_A).populate()
    
    gemmi_position = gemmi.Position(atomic_position[0], atomic_position[1], atomic_position[2])
    
    neighbors = ns.find_atoms(gemmi_position, '\0', radius=window_size_A)
    atoms = [gemmi_structure[0][x.chain_idx][x.residue_idx][x.atom_idx] for x in neighbors]
    number_of_neighbors = len(atoms)
    
    
    return number_of_neighbors

def get_atomic_bfactor_window(input_pdb, atomic_position, window_size_A):
    gemmi_structure = detect_pdb_input(input_pdb)
    
    # Neighbor Search initialize
    
    ns = gemmi.NeighborSearch(gemmi_structure[0], gemmi_structure.cell, window_size_A).populate()
    
    gemmi_position = gemmi.Position(atomic_position[0], atomic_position[1], atomic_position[2])
    
    neighbors = ns.find_atoms(gemmi_position, '\0', radius=window_size_A)
    atoms = [gemmi_structure[0][x.chain_idx][x.residue_idx][x.atom_idx] for x in neighbors]
    atomic_bfactor_list = np.array([x.b_iso for x in atoms])
    
    average_atomic_bfactor = atomic_bfactor_list.mean()
    
    return average_atomic_bfactor

def find_radius_of_gyration(input_pdb):
    
    structure = detect_pdb_input(input_pdb)
    
    num_atoms = structure[0].count_atom_sites()
    com = structure[0].calculate_center_of_mass()
    distances = []
    for model in structure:
        for chain in model:
            for res in chain:
                ca = res.get_ca()
                if ca is not None:
                    distances.append(com.dist(ca.pos))
    
    np_distance = np.array(distances)
    
    Rg = np.sum(np_distance**2)/num_atoms
    
    return Rg


# -*- coding: utf-8 -*-
"""

"""
import numpy as np
import gemmi
import mrcfile
import matplotlib.pyplot as plt

#%%


def plot_mapContour_model(threshold=0, modelfile="", gemmi_strct="", mdlidx=0, mapfile="", gemmi_map="", projection_orientation='z'):
    """
    Function for plotting density map as contour plot with model overlay.
    

    Parameters
    ----------
    threshold : float, optional
        threshold value for plotting. The default is 0.
    modelfile : str, optional
        pdb file to use as model structure. Either modelfile or gemmi_strct must be specified.
    gemmi_strct : gemmi.Structure, optional
        Gemmi structure object containing at least 1 model. Either modelfile or gemmi_strct must be specified.
    mdlidx : int, optional
        if gemmi_strct is specified with more than one model, this index can be set to plot a specific model or set to a iterable sequence to plot multiple models.
    mapfile : str, optional
        filename of the target map to be plotted. Either mapfile or mrcfile_map must be specified.
    mrcfile_map : MrcFile, optional
        mrcFile object of the target map.  Either mapfile or mrcfile_map must be specified.
    projection_orientation : str, optional
        axis along which to project the target map, can be 'x', 'y' or 'z'. The default is 'z'.


    Returns
    -------
    None.

    """
    
    if mapfile!="":
        # target_map = mrcfile.open(mapfile)
        target_map = gemmi.read_ccp4_map(mapfile)
    elif gemmi_map!="":
        target_map = gemmi_map
    else:
        print("must specify target map as either path to .mrc file or as mrcfile object")
        return 0
    # voxelsize = target_map.header.cella.x / target_map.header.nx
    voxelsize = target_map.grid.spacing[0]
    
    if projection_orientation=='x':
        projection_axis = 0
        plot_axes = [1,2]
    elif projection_orientation=='y':
        projection_axis = 1
        plot_axes = [0,2]
    elif projection_orientation=='z':
        projection_axis = 2
        plot_axes = [0,1]
    else:
        print("please specify valid projection orientation ('x','y','z')")
        return 0
    
    if modelfile!="":
        gemmi_strct = gemmi.read_structure(modelfile)
    elif gemmi_strct=="":
        print("must specify either path to .pdb file or a gemmi.Structure object")
        return 0
        
    print('gotten this far')
    T = np.copy(target_map.grid) # mrcfile.data object is read only
    T[T<threshold] = np.nan
    print(np.nanmax(T))    
    x = np.linspace(0, target_map.grid.unit_cell.a, target_map.grid.nu)
    y = np.linspace(0, target_map.grid.unit_cell.b, target_map.grid.nv)
    z = np.linspace(0, target_map.grid.unit_cell.c, target_map.grid.nw)
    if projection_orientation=='x':
        X1,X2 = np.meshgrid(y,z)
    elif projection_orientation=='y':
        X1,X2 = np.meshgrid(x,z)
    else:
        X1,X2 = np.meshgrid(x,y)
    
    # prepare plot figure
    projection = np.nanmean(T, axis=projection_axis).T
    # projection = np.mean(target_map.data,axis=2)
    fig, ax = plt.subplots()
    # im = ax.imshow(projection, interpolation='spline16')
    cntr = ax.contourf(X1,X2,projection, levels = 32, alpha=0.6)
    cbar = fig.colorbar(cntr)
    ax.axis('off')
    ax.set_aspect('equal')
    
    # get number of backbone atoms, assuming it's the same for all models
    numAtoms = 0
    for chn in gemmi_strct[0]:
        for res in chn:
            for atm in res:
                if atm.name == 'C' or atm.name=='N' or atm.name=='CA':
                    numAtoms += 1
    
    if hasattr(mdlidx,'__iter__'):
        legnames = []
        for i in mdlidx:
            # find positions of all backbone atoms
            plot_pos = np.zeros((numAtoms, 3))
            atmidx = 0
            for chn in gemmi_strct[i]:
                for res in chn:
                    for atm in res:
                        if atm.name == 'C' or atm.name=='N' or atm.name=='CA':
                            plot_pos[atmidx,:] = np.array([atm.pos.x, atm.pos.y, atm.pos.z])
                            atmidx+=1
        
            ax.plot(plot_pos[:,plot_axes[0]], plot_pos[:,plot_axes[1]],
                    marker='o',
                    markeredgecolor='black',
                    markersize=3,
                    linewidth=1.5) 
            legnames.append('timepoint = '+str(i))
        ax.legend(legnames)
        
    else:       
        # find positions of all backbone atoms
        plot_pos = np.zeros((numAtoms, 3))
        atmidx = 0
        for chn in gemmi_strct[mdlidx]:
            for res in chn:
                for atm in res:
                    if atm.name == 'C' or atm.name=='N' or atm.name=='CA':
                        plot_pos[atmidx,:] = np.array([atm.pos.x, atm.pos.y, atm.pos.z])
                        atmidx+=1
    
        ax.plot(plot_pos[:,plot_axes[0]], plot_pos[:,plot_axes[1]],
                marker='o',
                markerfacecolor='red',
                markeredgecolor='black',
                markersize=3,
                linewidth=1.5,
                color='red')    
        
    # x1lim = (plot_pos[:,plot_axes[0]].min() * 0.8, plot_pos[:,plot_axes[0]].max() * 1.2)
    # x2lim = (plot_pos[:,plot_axes[1]].min() * 0.8, plot_pos[:,plot_axes[1]].max() * 1.2)
    # ax.set_xlim(x1lim)
    # ax.set_ylim(x2lim)
    return fig, ax
            


## Script to measure surface area of an EM-map at a given threshold

def surface_area_at_threshold(emmap, apix, reference_threshold):
    import numpy as np

    from emmer.ndimage.contour.contour_utils import mesh_surface_area
    
    binarised_emmap = (emmap>=reference_threshold).astype(np.int_)
    surface_area = mesh_surface_area(binarised_emmap, 0.9999999, apix)
    return surface_area
